from typing import Protocol
from typing import runtime_checkable

from termcolor import cprint


@runtime_checkable
class DspToolsWarning(Protocol):
    """Protocol for warnings that implement a custom showwarnings() function"""

    @classmethod
    def showwarning(cls, message: str) -> None:
        """Functionality that should be executed when a warning of this class is emitted"""


class DspToolsUserWarning(Warning):
    """Class for general user-facing warnings"""

    @classmethod
    def showwarning(cls, message: str) -> None:
        """Print the warning, without context"""
        cprint(f"WARNING: {message}", color="red", attrs=["bold"])


class DspToolsFutureWarning(FutureWarning):
    """Class for user-facing deprecation warnings"""

    @classmethod
    def showwarning(cls, message: str) -> None:
        """Print the warning, without context"""
        cprint(f"DEPRECATION WARNING: {message}", color="red", attrs=["bold"])



import pytest
from rdflib import RDF
from rdflib import BNode
from rdflib import Graph
from rdflib import Literal
from rdflib import Namespace
from rdflib import URIRef

from dsp_tools.commands.xmlupload.models.serialise.jsonld_serialiser import _frame_property
from dsp_tools.commands.xmlupload.models.serialise.jsonld_serialiser import serialise_property_graph

KNORA_API = Namespace("http://api.knora.org/ontology/knora-api/v2#")
MY_ONTO = Namespace("http://0.0.0.0:3333/ontology/0009/myonto/v2#")


@pytest.fixture
def int_value() -> Graph:
    g = Graph()
    res_bn = BNode()
    val_bn = BNode()
    g.add((res_bn, MY_ONTO.hasInteger, val_bn))
    g.add((val_bn, RDF.type, KNORA_API.IntValue))
    g.add((val_bn, KNORA_API.intValueAsInt, Literal(1)))
    return g


@pytest.fixture
def int_value_with_comment() -> Graph:
    g = Graph()
    res_bn = BNode()
    val_bn = BNode()
    g.add((res_bn, MY_ONTO.hasInteger, val_bn))
    g.add((val_bn, RDF.type, KNORA_API.IntValue))
    g.add((val_bn, KNORA_API.intValueAsInt, Literal(2)))
    g.add((val_bn, KNORA_API.hasPermissions, Literal("CR knora-admin:Creator")))
    g.add((val_bn, KNORA_API.valueHasComment, Literal("this is the number 2")))
    return g


@pytest.fixture
def two_int_values() -> Graph:
    g = Graph()
    res_bn = BNode()
    val_bn1 = BNode()
    g.add((res_bn, MY_ONTO.hasInteger, val_bn1))
    g.add((val_bn1, RDF.type, KNORA_API.IntValue))
    g.add((val_bn1, KNORA_API.intValueAsInt, Literal(1)))
    val_bn2 = BNode()
    g.add((res_bn, MY_ONTO.hasInteger, val_bn2))
    g.add((val_bn2, RDF.type, KNORA_API.IntValue))
    g.add((val_bn2, KNORA_API.intValueAsInt, Literal(2)))
    return g


class TestSerialiseProperty:
    def test_one_value(self, int_value: Graph) -> None:
        expected = {
            "http://0.0.0.0:3333/ontology/0009/myonto/v2#hasInteger": {
                "@type": "http://api.knora.org/ontology/knora-api/v2#IntValue",
                "http://api.knora.org/ontology/knora-api/v2#intValueAsInt": 1,
            }
        }
        result = serialise_property_graph(int_value, MY_ONTO.hasInteger)
        assert result == expected

    def test_value_with_comments(self, int_value_with_comment: Graph) -> None:
        expected = {
            "http://0.0.0.0:3333/ontology/0009/myonto/v2#hasInteger": {
                "@type": "http://api.knora.org/ontology/knora-api/v2#IntValue",
                "http://api.knora.org/ontology/knora-api/v2#hasPermissions": "CR knora-admin:Creator",
                "http://api.knora.org/ontology/knora-api/v2#valueHasComment": "this is the number 2",
                "http://api.knora.org/ontology/knora-api/v2#intValueAsInt": 2,
            }
        }
        result = serialise_property_graph(int_value_with_comment, MY_ONTO.hasInteger)
        assert result == expected

    def test_several_values(self, two_int_values: Graph) -> None:
        expected = {
            "http://0.0.0.0:3333/ontology/0009/myonto/v2#hasInteger": [
                {
                    "@type": "http://api.knora.org/ontology/knora-api/v2#IntValue",
                    "http://api.knora.org/ontology/knora-api/v2#intValueAsInt": 1,
                },
                {
                    "@type": "http://api.knora.org/ontology/knora-api/v2#IntValue",
                    "http://api.knora.org/ontology/knora-api/v2#intValueAsInt": 2,
                },
            ],
        }

        result = serialise_property_graph(two_int_values, MY_ONTO.hasInteger)
        assert result == expected


def test_frame_property() -> None:
    json_graph = [
        {
            "@id": "_:res_id",
            "http://0.0.0.0:3333/ontology/0009/myonto2/v2#hasInteger": [{"@id": "_:1"}],
            "http://0.0.0.0:3333/ontology/0009/myonto/v2#hasGeoname": [{"@id": "_:2"}, {"@id": "_:3"}],
        },
        {
            "@id": "_:1",
            "@type": ["http://api.knora.org/ontology/knora-api/v2#IntValue"],
            "http://api.knora.org/ontology/knora-api/v2#intValueAsInt": [{"@value": 1}],
        },
        {
            "@id": "_:2",
            "@type": ["http://api.knora.org/ontology/knora-api/v2#GeonameValue"],
            "http://api.knora.org/ontology/knora-api/v2#geonameValueAsGeonameCode": [{"@value": "geo2"}],
        },
        {
            "@id": "_:3",
            "@type": ["http://api.knora.org/ontology/knora-api/v2#GeonameValue"],
            "http://api.knora.org/ontology/knora-api/v2#geonameValueAsGeonameCode": [{"@value": "geo3"}],
        },
    ]
    expected = {
        "http://0.0.0.0:3333/ontology/0009/myonto2/v2#hasInteger": {
            "@type": "http://api.knora.org/ontology/knora-api/v2#IntValue",
            "http://api.knora.org/ontology/knora-api/v2#intValueAsInt": 1,
        },
        "http://0.0.0.0:3333/ontology/0009/myonto/v2#hasGeoname": [
            {
                "@type": "http://api.knora.org/ontology/knora-api/v2#GeonameValue",
                "http://api.knora.org/ontology/knora-api/v2#geonameValueAsGeonameCode": "geo2",
            },
            {
                "@type": "http://api.knora.org/ontology/knora-api/v2#GeonameValue",
                "http://api.knora.org/ontology/knora-api/v2#geonameValueAsGeonameCode": "geo3",
            },
        ],
    }
    res = _frame_property(json_graph, URIRef("http://0.0.0.0:3333/ontology/0009/myonto2/v2#hasInteger"))
    assert res == expected


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

import uuid
from dataclasses import dataclass
from dataclasses import field


@dataclass(frozen=True)
class ResptrLink:
    """
    This class represents a direct link (resptr) between a starting resource and a target resource.

    Attributes:
        source_id: ID of the resource from which the link originates
        target_id: ID of the resource where the link points to
        link_uuid: identifier of this link
    """

    source_id: str
    target_id: str
    link_uuid: str = field(default_factory=lambda: str(uuid.uuid4()))

    @property
    def cost_links(self) -> float:
        """The cost of this outgoing is consistently 1"""
        return 1


@dataclass(frozen=True)
class XMLLink:
    """
    This class represents one or more links from a single starting resource to a set of target resources,
    where all target resources are linked to from a single text value of the starting resource.

    Attributes:
        source_id: ID of the resource from which the link(s) originate
        target_ids: IDs of the resources that are referenced in the text value
        link_uuid: identifier of this link
    """

    source_id: str
    target_ids: set[str]
    link_uuid: str = field(default_factory=lambda: str(uuid.uuid4()))

    @property
    def cost_links(self) -> float:
        """The cost of this outgoing link (1 / number of links in the XML text)"""
        return 1 / len(self.target_ids)


@dataclass(frozen=True)
class Edge:
    """
    This class represents an edge in the rustworkx graph.

    Attributes:
        source: rustworkx index of the resource from which the link originates
        target: rustworkx index of the resource where the link points to
        link_object: the link that connects the source with the target
    """

    source: int
    target: int
    link_object: ResptrLink | XMLLink

    def as_tuple(self) -> tuple[int, int, ResptrLink | XMLLink]:
        """Returns a representation of this edge as a tuple of the source index, target index and link object"""
        return self.source, self.target, self.link_object


@dataclass(frozen=True)
class Cost:
    """
    Attributes:
        source: rustworkx index of the resource from which the link originates
        target: rustworkx index of the resource where the link points to
        node_value: cost-gain-ratio if this link is stashed
    """

    source: int
    target: int
    node_value: float



from collections import defaultdict

import pytest
from lxml import etree
from pytest_unordered import unordered

from dsp_tools.commands.xmlupload.check_consistency_with_ontology import _check_all_classes_and_properties_in_onto
from dsp_tools.commands.xmlupload.check_consistency_with_ontology import _check_correctness_all_text_value_encodings
from dsp_tools.commands.xmlupload.check_consistency_with_ontology import _check_correctness_of_one_prop
from dsp_tools.commands.xmlupload.check_consistency_with_ontology import _find_all_class_types_in_onto
from dsp_tools.commands.xmlupload.check_consistency_with_ontology import _find_all_properties_in_onto
from dsp_tools.commands.xmlupload.check_consistency_with_ontology import _find_one_class_type_in_onto
from dsp_tools.commands.xmlupload.check_consistency_with_ontology import _find_one_property_in_onto
from dsp_tools.commands.xmlupload.check_consistency_with_ontology import _get_all_class_types_and_ids_from_data
from dsp_tools.commands.xmlupload.check_consistency_with_ontology import _get_all_classes_and_properties_from_data
from dsp_tools.commands.xmlupload.check_consistency_with_ontology import _get_all_ids_and_props_and_encodings_from_root
from dsp_tools.commands.xmlupload.check_consistency_with_ontology import (
    _get_all_property_names_and_resource_ids_one_resource,
)
from dsp_tools.commands.xmlupload.check_consistency_with_ontology import (
    _get_id_and_props_and_encodings_from_one_resource,
)
from dsp_tools.commands.xmlupload.check_consistency_with_ontology import _get_prop_and_encoding_from_one_property
from dsp_tools.commands.xmlupload.check_consistency_with_ontology import (
    _get_separate_prefix_and_iri_from_onto_prop_or_cls,
)
from dsp_tools.commands.xmlupload.models.ontology_lookup_models import OntoInfo
from dsp_tools.commands.xmlupload.models.ontology_lookup_models import ProjectOntosInformation
from dsp_tools.commands.xmlupload.models.ontology_lookup_models import PropertyTextValueTypes
from dsp_tools.commands.xmlupload.models.ontology_lookup_models import TextValueData


class TestFindClassType:
    def test_no_knora_prefix(self) -> None:
        onto_lookup = ProjectOntosInformation(
            default_ontology_prefix="test",
            onto_lookup={
                "test": OntoInfo(classes=["classA", "classB"], properties=[]),
                "knora-api": OntoInfo(classes=["knoraClassA"], properties=[]),
            },
        )
        assert not _find_one_class_type_in_onto("knoraClassA", onto_lookup)

    def test_knora_prefix(self) -> None:
        onto_lookup = ProjectOntosInformation(
            default_ontology_prefix="test",
            onto_lookup={
                "test": OntoInfo(classes=["classA", "classB"], properties=[]),
                "knora-api": OntoInfo(classes=["knoraClassA"], properties=[]),
            },
        )
        assert not _find_one_class_type_in_onto("knora-api:knoraClassA", onto_lookup)

    def test_no_default_prefix(self) -> None:
        onto_lookup = ProjectOntosInformation(
            default_ontology_prefix="test",
            onto_lookup={
                "test": OntoInfo(classes=["classA", "classB"], properties=[]),
                "knora-api": OntoInfo(classes=["knoraClassA"], properties=[]),
            },
        )
        assert not _find_one_class_type_in_onto(":classA", onto_lookup)

    def test_default_prefix(self) -> None:
        onto_lookup = ProjectOntosInformation(
            default_ontology_prefix="test",
            onto_lookup={
                "test": OntoInfo(classes=["classA", "classB"], properties=[]),
                "knora-api": OntoInfo(classes=["knoraClassA"], properties=[]),
            },
        )
        assert not _find_one_class_type_in_onto("test:classB", onto_lookup)

    def test_unknown_class(self) -> None:
        onto_lookup = ProjectOntosInformation(
            default_ontology_prefix="test",
            onto_lookup={
                "test": OntoInfo(classes=["classA", "classB"], properties=[]),
                "knora-api": OntoInfo(classes=["knoraClassA"], properties=[]),
            },
        )
        assert _find_one_class_type_in_onto("test:classC", onto_lookup) == "Invalid Class Type"

    def test_unknown_prefix(self) -> None:
        onto_lookup = ProjectOntosInformation(
            default_ontology_prefix="test",
            onto_lookup={
                "test": OntoInfo(classes=["classA", "classB"], properties=[]),
                "knora-api": OntoInfo(classes=["knoraClassA"], properties=[]),
            },
        )
        assert _find_one_class_type_in_onto("other:classC", onto_lookup) == "Unknown ontology prefix"

    def test_diagnose_all_classes_no_problems(self) -> None:
        onto_lookup = ProjectOntosInformation(
            default_ontology_prefix="test",
            onto_lookup={
                "test": OntoInfo(classes=["classA", "classB"], properties=[]),
                "knora-api": OntoInfo(classes=["knoraClassA"], properties=[]),
            },
        )
        test_classes = {"knora-api:knoraClassA": ["idA"], ":classB": ["idB"]}
        assert not _find_all_class_types_in_onto(test_classes, onto_lookup)

    def test_diagnose_all_classes_problems(self) -> None:
        onto_lookup = ProjectOntosInformation(
            default_ontology_prefix="test",
            onto_lookup={
                "test": OntoInfo(classes=["classA", "classB"], properties=[]),
                "knora-api": OntoInfo(classes=["knoraClassA"], properties=[]),
            },
        )
        test_classes = {"knoraClassA": ["idA"], ":classD": ["idD"]}
        assert _find_all_class_types_in_onto(test_classes, onto_lookup) == [(":classD", ["idD"], "Invalid Class Type")]


class TestFindProperties:
    def test_no_knora_prefix(self) -> None:
        onto_lookup = ProjectOntosInformation(
            default_ontology_prefix="test",
            onto_lookup={
                "test": OntoInfo(classes=[], properties=["propA", "propB"]),
                "knora-api": OntoInfo(classes=[], properties=["knoraPropA"]),
            },
        )
        assert not _find_one_property_in_onto("knoraPropA", onto_lookup)

    def test_knora_prefix(self) -> None:
        onto_lookup = ProjectOntosInformation(
            default_ontology_prefix="test",
            onto_lookup={
                "test": OntoInfo(classes=[], properties=["propA", "propB"]),
                "knora-api": OntoInfo(classes=[], properties=["knoraPropA"]),
            },
        )
        assert not _find_one_property_in_onto("knora-api:knoraPropA", onto_lookup)

    def test_no_default_prefix(self) -> None:
        onto_lookup = ProjectOntosInformation(
            default_ontology_prefix="test",
            onto_lookup={
                "test": OntoInfo(classes=[], properties=["propA", "propB"]),
                "knora-api": OntoInfo(classes=[], properties=["knoraPropA"]),
            },
        )
        assert not _find_one_property_in_onto(":propA", onto_lookup)

    def test_default_prefix(self) -> None:
        onto_lookup = ProjectOntosInformation(
            default_ontology_prefix="test",
            onto_lookup={
                "test": OntoInfo(classes=[], properties=["propA", "propB"]),
                "knora-api": OntoInfo(classes=[], properties=["knoraPropA"]),
            },
        )
        assert not _find_one_property_in_onto("test:propB", onto_lookup)

    def test_unknown_prefix(self) -> None:
        onto_lookup = ProjectOntosInformation(
            default_ontology_prefix="test",
            onto_lookup={
                "test": OntoInfo(classes=[], properties=["propA", "propB"]),
                "knora-api": OntoInfo(classes=[], properties=["knoraPropA"]),
            },
        )
        assert _find_one_property_in_onto("other:propB", onto_lookup) == "Unknown ontology prefix"

    def test_unknown_property(self) -> None:
        onto_lookup = ProjectOntosInformation(
            default_ontology_prefix="test",
            onto_lookup={
                "test": OntoInfo(classes=[], properties=["propA", "propB"]),
                "knora-api": OntoInfo(classes=[], properties=["knoraPropA"]),
            },
        )
        assert _find_one_property_in_onto("test:propC", onto_lookup) == "Invalid Property"

    def test_diagnose_all_properties_problems(self) -> None:
        onto_lookup = ProjectOntosInformation(
            default_ontology_prefix="test",
            onto_lookup={
                "test": OntoInfo(classes=[], properties=["propA", "propB"]),
                "knora-api": OntoInfo(classes=[], properties=["knoraPropA"]),
            },
        )
        test_properties = {"test:propB": ["idB"], "other:propB": ["idB"], "test:propD": ["idD"]}
        expected = [("other:propB", ["idB"], "Unknown ontology prefix"), ("test:propD", ["idD"], "Invalid Property")]
        assert unordered(_find_all_properties_in_onto(test_properties, onto_lookup)) == expected

    def test_diagnose_all_properties_no_problems(self) -> None:
        onto_lookup = ProjectOntosInformation(
            default_ontology_prefix="test",
            onto_lookup={
                "test": OntoInfo(classes=[], properties=["propA", "propB"]),
                "knora-api": OntoInfo(classes=[], properties=["knoraPropA"]),
            },
        )
        test_properties = {"test:propB": ["idB"], "knoraPropA": ["idA"]}
        assert not unordered(_find_all_properties_in_onto(test_properties, onto_lookup))


def test_check_all_classes_and_properties_in_onto() -> None:
    onto_lookup = ProjectOntosInformation(
        default_ontology_prefix="test",
        onto_lookup={
            "test": OntoInfo(classes=["classA", "classB"], properties=["propA", "propB"]),
            "knora-api": OntoInfo(classes=["knoraClassA"], properties=["knoraPropA"]),
        },
    )
    classes = {"knoraClassA": ["idA"]}
    properties = {"knora-api:knoraPropA": ["idA"]}
    res_msg = _check_all_classes_and_properties_in_onto(classes, properties, onto_lookup)
    assert not res_msg


def test_check_all_classes_and_properties_in_onto_problem() -> None:
    onto_lookup = ProjectOntosInformation(
        default_ontology_prefix="test",
        onto_lookup={
            "test": OntoInfo(classes=["classA", "classB"], properties=["propA", "propB"]),
            "knora-api": OntoInfo(classes=["knoraClassA"], properties=["knoraPropA"]),
        },
    )
    classes = {"knora": ["idA"]}
    properties = {"knora-api:knoraPropA": ["idA"]}
    expected_msg = (
        "\nSome property and/or class type(s) used in the XML are unknown.\n"
        "The ontologies for your project on the server are:\n"
        "    - test\n"
        "    - knora-api"
        "\n\n"
        "The following resource(s) have an invalid resource type:\n\n"
        "    Resource Type: 'knora'\n"
        "    Problem: 'Invalid Class Type'\n"
        "    Resource ID(s):\n"
        "    - idA"
        "\n\n"
    )
    res_msg = _check_all_classes_and_properties_in_onto(classes, properties, onto_lookup)
    assert res_msg == expected_msg


def test_get_prefix_and_prop_cls_identifier() -> None:
    assert _get_separate_prefix_and_iri_from_onto_prop_or_cls(":propA", "test") == ("test", "propA")
    assert _get_separate_prefix_and_iri_from_onto_prop_or_cls("knora-api:knoraPropA", "test") == (
        "knora-api",
        "knoraPropA",
    )
    assert _get_separate_prefix_and_iri_from_onto_prop_or_cls("knoraPropB", "test") == ("knora-api", "knoraPropB")
    assert _get_separate_prefix_and_iri_from_onto_prop_or_cls("test:propA", "test") == ("test", "propA")


def test_get_prefix_and_prop_cls_identifier_error() -> None:
    assert _get_separate_prefix_and_iri_from_onto_prop_or_cls("123654/", "test") == (None, None)


def test_get_all_class_types_and_ids_from_data() -> None:
    test_ele = etree.fromstring(
        """<knora shortcode="4123" default-ontology="test">
                <resource label="resA" restype=":TestThing1" id="resA" permissions="res-default"/>
                <resource label="resB" restype=":TestThing2" id="resB" permissions="res-default"/>
                <resource label="resC" restype=":TestThing2" id="resC" permissions="res-default"/>
        </knora>"""
    )
    expected_classes = {":TestThing1": ["resA"], ":TestThing2": ["resB", "resC"]}
    res_classes = _get_all_class_types_and_ids_from_data(test_ele)
    assert res_classes.keys() == expected_classes.keys()
    for k, v in res_classes.items():
        assert unordered(v) == expected_classes[k]


def test_get_all_property_names_and_resource_ids_one_resource() -> None:
    test_ele = etree.fromstring(
        """<resource label="resC" restype=":TestThing2" id="resC" permissions="res-default">
                <resptr-prop name=":hasResource2">
                    <resptr permissions="prop-default">resB</resptr>
                </resptr-prop>
                <resptr-prop name=":hasResource3">
                    <resptr permissions="prop-default">resA</resptr>
                </resptr-prop>
            </resource>"""
    )
    expected = {"a": ["a"], ":hasResource2": ["resB", "resC"], ":hasResource3": ["resC"]}
    res_dic = _get_all_property_names_and_resource_ids_one_resource(
        test_ele, defaultdict(list, {"a": ["a"], ":hasResource2": ["resB"]})
    )
    assert res_dic.keys() == expected.keys()
    for k, v in res_dic.items():
        assert unordered(v) == expected[k]


def test_get_all_classes_and_properties_from_data() -> None:
    test_ele = etree.fromstring(
        """<knora shortcode="4123" default-ontology="test">
                <resource label="resA" restype=":TestThing1" id="resA" permissions="res-default">
                    <resptr-prop name=":hasResource1">
                        <resptr permissions="prop-default">resB</resptr>
                    </resptr-prop>
                </resource>
                <resource label="resB" restype=":TestThing2" id="resB" permissions="res-default">
                </resource>
                <resource label="resC" restype=":TestThing2" id="resC" permissions="res-default">
                    <resptr-prop name=":hasResource2">
                        <resptr permissions="prop-default">resB</resptr>
                    </resptr-prop>
                    <resptr-prop name=":hasResource3">
                        <resptr permissions="prop-default">resA</resptr>
                    </resptr-prop>
                </resource>
        </knora>"""
    )
    expected_classes = {":TestThing1": ["resA"], ":TestThing2": ["resB", "resC"]}
    expected_properties = {
        ":hasResource1": ["resA"],
        ":hasResource2": ["resC"],
        ":hasResource3": ["resC"],
    }
    res_classes, res_properties = _get_all_classes_and_properties_from_data(test_ele)
    assert res_classes.keys() == expected_classes.keys()
    for k, v in res_classes.items():
        assert unordered(v) == expected_classes[k]
    assert res_properties.keys() == expected_properties.keys()
    for k, v in res_properties.items():
        assert unordered(v) == expected_properties[k]


def test_get_id_and_props_and_encodings_from_one_resource_no_text() -> None:
    test_props = etree.fromstring(
        """
        <resource label="First Testthing"
              restype=":TestThing"
              id="test_thing_1"
              permissions="res-default">
            <uri-prop name=":hasUri">
                <uri permissions="prop-default">https://dasch.swiss</uri>
            </uri-prop>
            <boolean-prop name=":hasBoolean">
                <boolean permissions="prop-default">true</boolean>
            </boolean-prop>
        </resource>
        """
    )
    res = _get_id_and_props_and_encodings_from_one_resource(test_props)
    assert not res


def test_get_id_and_props_and_encodings_from_one_resource_richtext() -> None:
    test_props = etree.fromstring(
        """
        <resource label="First Testthing"
              restype=":TestThing"
              id="test_thing_1"
              permissions="res-default">
            <uri-prop name=":hasUri">
                <uri permissions="prop-default">https://dasch.swiss</uri>
            </uri-prop>
            <boolean-prop name=":hasBoolean">
                <boolean permissions="prop-default">true</boolean>
            </boolean-prop>
            <text-prop name=":hasRichtext">
                <text encoding="utf8">Text</text>
            </text-prop>
        </resource>
        """
    )
    res = _get_id_and_props_and_encodings_from_one_resource(test_props)[0]
    assert res.resource_id == "test_thing_1"
    assert res.property_name == ":hasRichtext"
    assert res.encoding == "utf8"


def test_get_id_and_props_and_encodings_from_one_resource_two_text_props() -> None:
    test_props = etree.fromstring(
        """
        <resource label="First Testthing"
              restype=":TestThing"
              id="test_thing_1"
              permissions="res-default">
            <uri-prop name=":hasUri">
                <uri permissions="prop-default">https://dasch.swiss</uri>
            </uri-prop>
            <boolean-prop name=":hasBoolean">
                <boolean permissions="prop-default">true</boolean>
            </boolean-prop>
            <text-prop name=":hasRichtext">
                <text encoding="xml">Text</text>
            </text-prop>
            <text-prop name=":hasSimpleText">
                <text encoding="utf8">Text</text>
            </text-prop>
        </resource>
        """
    )
    res = _get_id_and_props_and_encodings_from_one_resource(test_props)
    assert res[0].resource_id == "test_thing_1"
    assert res[0].property_name == ":hasRichtext"
    assert res[0].encoding == "xml"
    assert res[1].resource_id == "test_thing_1"
    assert res[1].property_name == ":hasSimpleText"
    assert res[1].encoding == "utf8"


class TestGetEncodingOneProperty:
    def test_richtext_several_text_ele(self) -> None:
        test_prop = etree.fromstring(
            """
            <text-prop name=":hasRichtext">
                <text encoding="xml">&lt;</text>
                <text encoding="xml" permissions="prop-default">
                    This text contains links to all resources:
                    <a class="salsah-link" href="IRI:test_thing_0:IRI">test_thing_0</a>
                </text>
                <text encoding="xml">Text with an external link: <a href="https://www.google.com/">Google</a></text>
            </text-prop>
            """
        )
        res_info = _get_prop_and_encoding_from_one_property("id", ":resType", test_prop)
        assert res_info.resource_id == "id"
        assert res_info.res_type == ":resType"
        assert res_info.property_name == ":hasRichtext"
        assert res_info.encoding == "xml"

    def test_simple_several_text_ele(self) -> None:
        test_prop = etree.fromstring(
            """
            <text-prop name=":hasRichtext">
                <text encoding="utf8">Text</text>
                <text encoding="utf8">Text</text>
            </text-prop>
            """
        )
        res_info = _get_prop_and_encoding_from_one_property("id", ":resType", test_prop)
        assert res_info.resource_id == "id"
        assert res_info.res_type == ":resType"
        assert res_info.property_name == ":hasRichtext"
        assert res_info.encoding == "utf8"

    def test_simple_one_text_ele(self) -> None:
        test_prop = etree.fromstring(
            """
            <text-prop name=":hasRichtext">
                <text encoding="utf8">Text</text>
            </text-prop>
            """
        )
        res_info = _get_prop_and_encoding_from_one_property("id", ":resType", test_prop)
        assert res_info.resource_id == "id"
        assert res_info.res_type == ":resType"
        assert res_info.property_name == ":hasRichtext"
        assert res_info.encoding == "utf8"


def test_get_all_ids_and_props_and_encodings_from_root_no_text() -> None:
    test_ele = etree.fromstring(
        """<knora>
                <resource label="resA" restype=":TestThing1" id="resA" permissions="res-default">
                    <resptr-prop name=":hasResource1">
                        <resptr permissions="prop-default">resB</resptr>
                    </resptr-prop>
                </resource>
                <resource label="resB" restype=":TestThing2" id="resB" permissions="res-default">
                </resource>
                <resource label="resC" restype=":TestThing2" id="resC" permissions="res-default">
                    <resptr-prop name=":hasResource2">
                        <resptr permissions="prop-default">resB</resptr>
                    </resptr-prop>
                    <resptr-prop name=":hasResource3">
                        <resptr permissions="prop-default">resA</resptr>
                    </resptr-prop>
                </resource>
        </knora>"""
    )
    res = _get_all_ids_and_props_and_encodings_from_root(test_ele)
    assert res == []


def test_get_all_ids_and_props_and_encodings_from_root_with_text() -> None:
    test_ele = etree.fromstring(
        """<knora>
                <resource label="First Testthing"
                      restype=":TestThing"
                      id="test_thing_1"
                      permissions="res-default">
                    <uri-prop name=":hasUri">
                        <uri permissions="prop-default">https://dasch.swiss</uri>
                    </uri-prop>
                    <text-prop name=":hasRichtext">
                        <text encoding="xml">Text</text>
                    </text-prop>
                </resource>
                <resource label="resB" restype=":TestThing2" id="resB" permissions="res-default">
                    <text-prop name=":hasSimpleText">
                        <text encoding="utf8">Text</text>
                    </text-prop>
                </resource>
                <resource label="resC" restype=":TestThing2" id="resC" permissions="res-default">
                    <resptr-prop name=":hasResource2">
                        <resptr permissions="prop-default">resB</resptr>
                    </resptr-prop>
                    <text-prop name=":hasSimpleText">
                        <text encoding="utf8">Text</text>
                    </text-prop>
                </resource>
                <resource label="resC" restype=":TestThing2" id="resD" permissions="res-default">
                    <resptr-prop name=":hasResource2">
                        <resptr permissions="prop-default">resB</resptr>
                    </resptr-prop>
                </resource>
        </knora>"""
    )
    res = _get_all_ids_and_props_and_encodings_from_root(test_ele)
    assert res[0].resource_id == "test_thing_1"
    assert res[0].property_name == ":hasRichtext"
    assert res[0].encoding == "xml"
    assert res[1].resource_id == "resB"
    assert res[1].property_name == ":hasSimpleText"
    assert res[1].encoding == "utf8"
    assert res[2].resource_id == "resC"
    assert res[2].property_name == ":hasSimpleText"
    assert res[2].encoding == "utf8"


class TestCheckCorrectnessOneProp:
    def test_utf_simple_correct(self) -> None:
        test_val = TextValueData("id", ":restype", ":prop", "utf8")
        test_lookup = PropertyTextValueTypes(set(), {":prop"})
        assert _check_correctness_of_one_prop(test_val, test_lookup) is True

    def test_utf_simple_wrong(self) -> None:
        test_val = TextValueData("id", ":restype", ":prop", "utf8")
        test_lookup = PropertyTextValueTypes({":prop"}, set())
        assert _check_correctness_of_one_prop(test_val, test_lookup) is False

    def test_xml_correct(self) -> None:
        test_val = TextValueData("id", ":restype", ":prop", "xml")
        test_lookup = PropertyTextValueTypes({":prop"}, set())
        assert _check_correctness_of_one_prop(test_val, test_lookup) is True

    def test_xml_wrong(self) -> None:
        test_val = TextValueData("id", ":restype", ":prop", "xml")
        test_lookup = PropertyTextValueTypes(set(), set())
        assert _check_correctness_of_one_prop(test_val, test_lookup) is False


def test_analyse_all_text_value_encodings_are_correct_all_good() -> None:
    test_ele = etree.fromstring(
        """<knora>
                <resource label="First Testthing"
                      restype=":TestThing"
                      id="test_thing_1"
                      permissions="res-default">
                    <uri-prop name=":hasUri">
                        <uri permissions="prop-default">https://dasch.swiss</uri>
                    </uri-prop>
                    <text-prop name=":hasRichtext">
                        <text encoding="xml">Text</text>
                    </text-prop>
                </resource>
                <resource label="resB" restype=":TestThing2" id="resB" permissions="res-default">
                    <text-prop name=":hasSimpleText">
                        <text encoding="utf8">Text</text>
                    </text-prop>
                </resource>
                <resource label="resC" restype=":TestThing2" id="resC" permissions="res-default">
                    <resptr-prop name=":hasResource2">
                        <resptr permissions="prop-default">resB</resptr>
                    </resptr-prop>
                    <text-prop name=":hasSimpleText">
                        <text encoding="utf8">Text</text>
                        <text encoding="utf8">Text</text>
                    </text-prop>
                </resource>
                <resource label="resC" restype=":TestThing2" id="resD" permissions="res-default">
                    <resptr-prop name=":hasResource2">
                        <resptr permissions="prop-default">resB</resptr>
                    </resptr-prop>
                </resource>
        </knora>"""
    )
    test_lookup = PropertyTextValueTypes(
        formatted_text_props={":hasRichtext"}, unformatted_text_props={":hasSimpleText", ":hasText"}
    )
    res_msg = _check_correctness_all_text_value_encodings(test_ele, test_lookup)
    assert not res_msg


def test_analyse_all_text_value_encodings_are_correct_problems() -> None:
    test_ele = etree.fromstring(
        """<knora>
                <resource label="First Testthing"
                      restype=":TestThing"
                      id="test_thing_1"
                      permissions="res-default">
                    <uri-prop name=":hasUri">
                        <uri permissions="prop-default">https://dasch.swiss</uri>
                    </uri-prop>
                    <text-prop name=":hasText">
                        <text encoding="xml">Text</text>
                    </text-prop>
                </resource>
                <resource label="resB" restype=":TestThing2" id="resB" permissions="res-default">
                    <text-prop name=":hasSimpleText">
                        <text encoding="utf8">Text</text>
                    </text-prop>
                </resource>
                <resource label="resC" restype=":TestThing2" id="resC" permissions="res-default">
                    <resptr-prop name=":hasResource2">
                        <resptr permissions="prop-default">resB</resptr>
                    </resptr-prop>
                    <text-prop name=":hasRichtext">
                        <text encoding="utf8">Text</text>
                        <text encoding="utf8">Text</text>
                    </text-prop>
                </resource>
                <resource label="resC" restype=":TestThing2" id="resD" permissions="res-default">
                    <resptr-prop name=":hasResource2">
                        <resptr permissions="prop-default">resB</resptr>
                    </resptr-prop>
                </resource>
        </knora>"""
    )
    test_lookup = PropertyTextValueTypes(
        formatted_text_props={":hasRichtext"}, unformatted_text_props={":hasSimpleText", ":hasText"}
    )
    expected_msg = (
        "\nSome text encodings used in the XML data file are not conform with the gui_element "
        "specified in the JSON ontology.\n"
        "Please consult the ontology regarding the assigned gui_elements."
        "\n\n"
        "Resource ID: 'resC' | Resource Type: ':TestThing2'\n"
        "    - Property Name: ':hasRichtext' -> Encoding Used: 'utf8'"
        "\n\n"
        "Resource ID: 'test_thing_1' | Resource Type: ':TestThing'\n"
        "    - Property Name: ':hasText' -> Encoding Used: 'xml'"
    )
    res_msg = _check_correctness_all_text_value_encodings(test_ele, test_lookup)
    assert res_msg == expected_msg


if __name__ == "__main__":
    pytest.main([__file__])



import pytest
from rdflib import XSD
from rdflib import Literal
from rdflib import Namespace
from rdflib import URIRef

from dsp_tools.commands.xml_validate.models.data_rdf import BooleanValueRDF
from dsp_tools.commands.xml_validate.models.data_rdf import ColorValueRDF
from dsp_tools.commands.xml_validate.models.data_rdf import DateValueRDF
from dsp_tools.commands.xml_validate.models.data_rdf import DecimalValueRDF
from dsp_tools.commands.xml_validate.models.data_rdf import GeonameValueRDF
from dsp_tools.commands.xml_validate.models.data_rdf import IntValueRDF
from dsp_tools.commands.xml_validate.models.data_rdf import LinkValueRDF
from dsp_tools.commands.xml_validate.models.data_rdf import ListValueRDF
from dsp_tools.commands.xml_validate.models.data_rdf import ResourceRDF
from dsp_tools.commands.xml_validate.models.data_rdf import RichtextRDF
from dsp_tools.commands.xml_validate.models.data_rdf import SimpleTextRDF
from dsp_tools.commands.xml_validate.models.data_rdf import TimeValueRDF
from dsp_tools.commands.xml_validate.models.data_rdf import UriValueRDF

ONTO = Namespace("http://0.0.0.0:3333/ontology/9999/onto/v2#")
KNORA_API = Namespace("http://api.knora.org/ontology/knora-api/v2#")


@pytest.fixture
def rdf_resource() -> ResourceRDF:
    return ResourceRDF(
        res_iri=URIRef("id"), res_class=ONTO.ClassWithEverything, label=Literal("lbl", datatype=XSD.string)
    )


@pytest.fixture
def rdf_boolean_value_corr() -> BooleanValueRDF:
    return BooleanValueRDF(ONTO.testBoolean, Literal("false", datatype=XSD.boolean), URIRef("id"))


@pytest.fixture
def rdf_boolean_value_empty() -> BooleanValueRDF:
    return BooleanValueRDF(ONTO.testBoolean, Literal("", datatype=XSD.string), URIRef("id"))


@pytest.fixture
def rdf_color_value_corr() -> ColorValueRDF:
    return ColorValueRDF(ONTO.testColor, Literal("00ff00", datatype=XSD.string), URIRef("id"))


@pytest.fixture
def rdf_date_value_corr() -> DateValueRDF:
    return DateValueRDF(
        ONTO.testSubDate1,
        Literal("JULIAN:BCE:0700:BCE:0600", datatype=XSD.string),
        URIRef("id"),
    )


@pytest.fixture
def rdf_decimal_value_corr() -> DecimalValueRDF:
    return DecimalValueRDF(ONTO.testDecimalSimpleText, Literal("1.2", datatype=XSD.decimal), URIRef("id"))


@pytest.fixture
def rdf_decimal_value_empty() -> DecimalValueRDF:
    return DecimalValueRDF(ONTO.testDecimalSimpleText, Literal("", datatype=XSD.string), URIRef("id"))


@pytest.fixture
def rdf_geoname_value_corr() -> GeonameValueRDF:
    return GeonameValueRDF(ONTO.testGeoname, Literal("1241345", datatype=XSD.integer), URIRef("id"))


@pytest.fixture
def rdf_integer_value_corr() -> IntValueRDF:
    return IntValueRDF(ONTO.testIntegerSimpleText, Literal("1", datatype=XSD.integer), URIRef("id"))


@pytest.fixture
def rdf_integer_value_empty() -> IntValueRDF:
    return IntValueRDF(ONTO.testIntegerSimpleText, Literal("", datatype=XSD.string), URIRef("id"))


@pytest.fixture
def rdf_link_value_corr() -> LinkValueRDF:
    return LinkValueRDF(ONTO.testHasLinkTo, URIRef("link-id"), URIRef("id"))


@pytest.fixture
def rdf_link_value_empty() -> LinkValueRDF:
    return LinkValueRDF(ONTO.testHasLinkTo, URIRef(""), URIRef("id"))


@pytest.fixture
def rdf_list_value_corr() -> ListValueRDF:
    return ListValueRDF(
        ONTO.testListProp,
        object_value=Literal("n1", datatype=XSD.string),
        list_name=Literal("onlyList", datatype=XSD.string),
        res_iri=URIRef("id"),
    )


@pytest.fixture
def rdf_list_value_wrong_node() -> ListValueRDF:
    return ListValueRDF(
        ONTO.testListProp,
        object_value=Literal("other", datatype=XSD.string),
        list_name=Literal("onlyList", datatype=XSD.string),
        res_iri=URIRef("id"),
    )


@pytest.fixture
def rdf_list_value_wrong_list() -> ListValueRDF:
    return ListValueRDF(
        ONTO.testListProp,
        object_value=Literal("n1", datatype=XSD.string),
        list_name=Literal("other", datatype=XSD.string),
        res_iri=URIRef("id"),
    )


@pytest.fixture
def rdf_text_richtext_value_corr() -> RichtextRDF:
    return RichtextRDF(ONTO.testRichtext, Literal("Text", datatype=XSD.string), URIRef("id"))


@pytest.fixture
def rdf_text_richtext_value_empty() -> RichtextRDF:
    return RichtextRDF(ONTO.testRichtext, Literal("", datatype=XSD.string), URIRef("id"))


@pytest.fixture
def rdf_text_simpletext_value_corr() -> SimpleTextRDF:
    return SimpleTextRDF(ONTO.testTextarea, Literal("Text", datatype=XSD.string), URIRef("id"))


@pytest.fixture
def rdf_time_value_corr() -> TimeValueRDF:
    return TimeValueRDF(
        ONTO.testTimeValue,
        Literal("2019-10-23T13:45:12.01-14:00", datatype=XSD.dateTimeStamp),
        URIRef("id"),
    )


@pytest.fixture
def rdf_time_value_empty() -> TimeValueRDF:
    return TimeValueRDF(ONTO.testTimeValue, Literal("", datatype=XSD.string), URIRef("id"))


@pytest.fixture
def rdf_uri_value_corr() -> UriValueRDF:
    return UriValueRDF(
        ONTO.testUriValue,
        Literal("https://dasch.swiss", datatype=XSD.anyURI),
        URIRef("id"),
    )


@pytest.fixture
def rdf_uri_value_empty() -> UriValueRDF:
    return UriValueRDF(ONTO.testUriValue, Literal("", datatype=XSD.string), URIRef("id"))


@pytest.fixture
def rdf_uri_value_wrong() -> UriValueRDF:
    return UriValueRDF(ONTO.testUriValue, Literal("with space", datatype=XSD.anyURI), URIRef("id"))



"""unit tests for the command line interface"""

import argparse
import unittest

import pytest

from dsp_tools.cli.entry_point import _derive_dsp_ingest_url
from dsp_tools.cli.entry_point import _get_canonical_server_and_dsp_ingest_url
from dsp_tools.models.exceptions import UserError

# ruff: noqa: PT009 (pytest-unittest-assertion) (remove this line when pytest is used instead of unittest)
# ruff: noqa: PT027 (pytest-unittest-raises-assertion) (remove this line when pytest is used instead of unittest)


class TestCLI(unittest.TestCase):
    """
    unit tests for the command line interface
    """

    default_dsp_api_url = "http://0.0.0.0:3333"
    default_dsp_ingest_url = "http://0.0.0.0:3340"
    root_user_email = "root@example.com"
    root_user_pw = "test"
    positive_testcases: dict[str, list[str]]
    negative_testcases: list[str]

    @classmethod
    def setUpClass(cls) -> None:
        """
        Populate the positive testcases and the negative testcases.
        Is executed once before the methods of this class are run.
        """
        cls.positive_testcases = {
            "https://0.0.0.0:3333/": [
                cls.default_dsp_api_url,
                cls.default_dsp_ingest_url,
            ],
            "0.0.0.0:3333": [
                cls.default_dsp_api_url,
                cls.default_dsp_ingest_url,
            ],
            "localhost:3333": [
                cls.default_dsp_api_url,
                cls.default_dsp_ingest_url,
            ],
            "https://admin.dasch.swiss": [
                "https://api.dasch.swiss",
                "https://ingest.dasch.swiss",
            ],
            "https://api.dasch.swiss": [
                "https://api.dasch.swiss",
                "https://ingest.dasch.swiss",
            ],
            "https://app.dasch.swiss": [
                "https://api.dasch.swiss",
                "https://ingest.dasch.swiss",
            ],
            "https://ingest.dasch.swiss": [
                "https://api.dasch.swiss",
                "https://ingest.dasch.swiss",
            ],
            "https://dasch.swiss": [
                "https://api.dasch.swiss",
                "https://ingest.dasch.swiss",
            ],
            "dasch.swiss": [
                "https://api.dasch.swiss",
                "https://ingest.dasch.swiss",
            ],
            "http://admin.test.dasch.swiss/": [
                "https://api.test.dasch.swiss",
                "https://ingest.test.dasch.swiss",
            ],
            "http://app.staging.dasch.swiss/": [
                "https://api.staging.dasch.swiss",
                "https://ingest.staging.dasch.swiss",
            ],
            "https://demo.dasch.swiss/": [
                "https://api.demo.dasch.swiss",
                "https://ingest.demo.dasch.swiss",
            ],
            "http://api.dev.dasch.swiss/": [
                "https://api.dev.dasch.swiss",
                "https://ingest.dev.dasch.swiss",
            ],
            "dev-02.dasch.swiss": [
                "https://api.dev-02.dasch.swiss",
                "https://ingest.dev-02.dasch.swiss",
            ],
            "082a-test-server.dasch.swiss": [
                "https://api.082a-test-server.dasch.swiss",
                "https://ingest.082a-test-server.dasch.swiss",
            ],
            "admin.08F4-test-server.dasch.swiss": [
                "https://api.08F4-test-server.dasch.swiss",
                "https://ingest.08F4-test-server.dasch.swiss",
            ],
            "app.08F4-test-server.dasch.swiss": [
                "https://api.08F4-test-server.dasch.swiss",
                "https://ingest.08F4-test-server.dasch.swiss",
            ],
            "ingest.08F4-test-server.dasch.swiss": [
                "https://api.08F4-test-server.dasch.swiss",
                "https://ingest.08F4-test-server.dasch.swiss",
            ],
            "not-yet-0826-test-server.dasch.swiss": [
                "https://api.not-yet-0826-test-server.dasch.swiss",
                "https://ingest.not-yet-0826-test-server.dasch.swiss",
            ],
            "https://admin.ls-prod-server.dasch.swiss": [
                "https://api.ls-prod-server.dasch.swiss",
                "https://ingest.ls-prod-server.dasch.swiss",
            ],
            "https://ls-test-server.dasch.swiss": [
                "https://api.ls-test-server.dasch.swiss",
                "https://ingest.ls-test-server.dasch.swiss",
            ],
        }
        cls.negative_testcases = [
            "https://0.0.0.0:1234",
            "https://api.unkown-host.ch",
        ]

    def test_derive_dsp_ingest_url_without_server(self) -> None:
        """
        If the argparse.Namespace does not contain a server,
        the function should return the same object.
        """
        args_without_server = argparse.Namespace(
            action="xmlupload",
            xmlfile="data.xml",
        )
        args_returned = _derive_dsp_ingest_url(
            parsed_arguments=args_without_server,
            default_dsp_api_url=self.default_dsp_api_url,
            default_dsp_ingest_url=self.default_dsp_ingest_url,
        )
        self.assertEqual(args_without_server, args_returned)

    def test_derive_dsp_ingest_url_with_server(self) -> None:
        """
        If the argparse.Namespace contains a server,
        the function should return a modified argparse.Namespace,
        with the correct DSP URL and Ingest URL.
        """
        args_with_server = argparse.Namespace(
            action="xmlupload",
            server="dasch.swiss",
            user="some.user@dasch.swiss",
            password="password",
            xmlfile="data.xml",
        )
        args_returned = _derive_dsp_ingest_url(
            parsed_arguments=args_with_server,
            default_dsp_api_url=self.default_dsp_api_url,
            default_dsp_ingest_url=self.default_dsp_ingest_url,
        )
        args_expected = argparse.Namespace(
            action="xmlupload",
            server="https://api.dasch.swiss",
            dsp_ingest_url="https://ingest.dasch.swiss",
            user="some.user@dasch.swiss",
            password="password",
            xmlfile="data.xml",
        )
        self.assertEqual(args_expected, args_returned)

    def test_get_canonical_server_and_dsp_ingest_url(self) -> None:
        """
        Test the method that canonicalizes the DSP URL and derives the SIPI URL from it.
        """
        for api_url_orig, expected in self.positive_testcases.items():
            api_url_expected, dsp_ingest_url_expected = expected
            api_url_returned, dsp_ingest_url_returned = _get_canonical_server_and_dsp_ingest_url(
                server=api_url_orig,
                default_dsp_api_url=self.default_dsp_api_url,
                default_dsp_ingest_url=self.default_dsp_ingest_url,
            )
            self.assertEqual(api_url_expected, api_url_returned)
            self.assertEqual(dsp_ingest_url_expected, dsp_ingest_url_returned)

        for invalid in self.negative_testcases:
            with self.assertRaisesRegex(UserError, r"Invalid DSP server URL"):
                _ = _get_canonical_server_and_dsp_ingest_url(
                    server=invalid,
                    default_dsp_api_url=self.default_dsp_api_url,
                    default_dsp_ingest_url=self.default_dsp_ingest_url,
                )


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from pathlib import Path

from lxml import etree

from dsp_tools.commands.excel2xml import append_permissions
from dsp_tools.commands.excel2xml import write_xml
from dsp_tools.xmllib.models.resource import Resource

XML_NAMESPACE_MAP = {None: "https://dasch.swiss/schema", "xsi": "http://www.w3.org/2001/XMLSchema-instance"}
DASCH_SCHEMA = "{https://dasch.swiss/schema}"


@dataclass
class XMLRoot:
    shortcode: str
    default_ontology: str
    resources: list[Resource] = field(default_factory=list)

    def make_root(self) -> etree._Element:
        schema_url = (
            "https://raw.githubusercontent.com/dasch-swiss/dsp-tools/main/src/dsp_tools/resources/schema/data.xsd"
        )
        schema_location_key = str(etree.QName("http://www.w3.org/2001/XMLSchema-instance", "schemaLocation"))
        schema_location_value = f"https://dasch.swiss/schema {schema_url}"
        return etree.Element(
            f"{DASCH_SCHEMA}knora",
            attrib={
                schema_location_key: schema_location_value,
                "shortcode": self.shortcode,
                "default-ontology": self.default_ontology,
            },
            nsmap=XML_NAMESPACE_MAP,
        )

    def add_resource(self, resource: Resource) -> XMLRoot:
        self.resources.append(resource)
        return self

    def add_resources(self, resources: list[Resource]) -> XMLRoot:
        self.resources.extend(resources)
        return self

    def add_resource_optional(self, resource: Resource | None) -> XMLRoot:
        if resource:
            self.resources.append(resource)
        return self

    def serialise(self) -> etree._Element:
        root = self.make_root()
        root = append_permissions(root)
        serialised_resources = [x.serialise() for x in self.resources]
        root.extend(serialised_resources)
        return root

    def write_file(self, filepath: str | Path) -> None:
        root = self.serialise()
        write_xml(root, filepath)



"""
This model implements the handling of resource classes. It contains two classes that work closely together:
    * "HasProperty" deals with the association of Property-instances with the Resource-instances. This association
      is done using the "cardinality"-clause
    * "ResourceClass" is the main class representing a DSP resource class.
"""

from __future__ import annotations

from enum import Enum
from typing import Any
from typing import Optional
from typing import Sequence
from typing import Union
from urllib.parse import quote_plus

import regex

from dsp_tools.commands.project.models.context import Context
from dsp_tools.commands.project.models.helpers import Cardinality
from dsp_tools.commands.project.models.model import Model
from dsp_tools.models.datetimestamp import DateTimeStamp
from dsp_tools.models.exceptions import BaseError
from dsp_tools.models.langstring import LangString
from dsp_tools.utils.connection import Connection


class HasProperty(Model):
    ROUTE: str = "/v2/ontologies/cardinalities"

    class Ptype(Enum):
        system = 1
        knora = 2
        other = 3

    _context: Context
    _ontology_id: str
    _property_id: str
    _resclass_id: str
    _cardinality: Cardinality
    _gui_order: int
    _ptype: Ptype

    def __init__(
        self,
        con: Connection,
        context: Context,
        ontology_id: Optional[str] = None,
        property_id: Optional[str] = None,
        resclass_id: Optional[str] = None,
        cardinality: Optional[Cardinality] = None,
        gui_order: Optional[int] = None,
        ptype: Optional[Ptype] = None,
    ):
        super().__init__(con)
        if not isinstance(context, Context):
            raise BaseError('"context"-parameter must be an instance of Context')
        self._context = context
        if ontology_id is not None:
            self._ontology_id = context.iri_from_prefix(ontology_id)
        else:
            self._ontology_id = None
        self._property_id = property_id
        self._resclass_id = resclass_id
        self._cardinality = cardinality
        self._gui_order = gui_order
        self._ptype = ptype
        self._changed = set()

    #
    # Here follows a list of getters/setters
    #
    @property
    def ontology_id(self) -> Optional[str]:
        return self._ontology_id

    @ontology_id.setter
    def ontology_id(self, value: Optional[str]) -> None:
        self._ontology_id = value

    @property
    def property_id(self) -> Optional[str]:
        return self._property_id

    @property
    def resclass_id(self) -> Optional[str]:
        return self._resclass_id

    @resclass_id.setter
    def resclass_id(self, value: Optional[str]) -> None:
        self._resclass_id = value

    @property
    def cardinality(self) -> Optional[Cardinality]:
        return self._cardinality

    @cardinality.setter
    def cardinality(self, value: Optional[Cardinality]) -> None:
        self._cardinality = value
        self._changed.add("cardinality")

    @property
    def gui_order(self) -> Optional[int]:
        return self._gui_order

    @gui_order.setter
    def gui_order(self, value: int) -> None:
        self._gui_order = value
        self._changed.add("gui_order")

    @property
    def ptype(self) -> Ptype:
        return self._ptype

    @classmethod
    def fromJsonObj(cls, con: Connection, context: Context, jsonld_obj: dict[str, Any]) -> tuple[str, HasProperty]:
        rdf_iri = context.prefix_from_iri("http://www.w3.org/1999/02/22-rdf-syntax-ns#")
        rdfs_iri = context.prefix_from_iri("http://www.w3.org/2000/01/rdf-schema#")
        owl_iri = context.prefix_from_iri("http://www.w3.org/2002/07/owl#")
        knora_api_iri = context.prefix_from_iri("http://api.knora.org/ontology/knora-api/v2#")
        salsah_gui_iri = context.prefix_from_iri("http://api.knora.org/ontology/salsah-gui/v2#")

        if jsonld_obj.get("@type") is None or jsonld_obj.get("@type") != owl_iri + ":Restriction":
            raise BaseError("Expected restriction type")

        cardinality = cls._fromJsonObj_get_cardinality(jsonld_obj, owl_iri)

        ontology_id, property_id, ptype = cls._fromJsonObj_get_prop_type_iri(
            context, jsonld_obj, knora_api_iri, owl_iri, rdf_iri, rdfs_iri
        )

        gui_order: int = None
        if jsonld_obj.get(salsah_gui_iri + ":guiOrder") is not None:
            gui_order = jsonld_obj[salsah_gui_iri + ":guiOrder"]
        return property_id, cls(
            con=con,
            context=context,
            ontology_id=ontology_id,
            property_id=property_id,
            cardinality=cardinality,
            gui_order=gui_order,
            ptype=ptype,
        )

    @staticmethod
    def _fromJsonObj_get_prop_type_iri(
        context: Context, jsonld_obj: dict[str, Any], knora_api_iri: str, owl_iri: str, rdf_iri: str, rdfs_iri: str
    ) -> tuple[str, str, HasProperty.Ptype]:
        property_id: str
        ptype: HasProperty.Ptype
        ontology_id: Optional[str] = None
        if jsonld_obj.get(owl_iri + ":onProperty") is None:
            raise BaseError("No property IRI given")
        p = jsonld_obj[owl_iri + ":onProperty"].get("@id")
        if p is None:
            raise BaseError("No property IRI given")
        pp = p.split(":")
        if pp[0] == rdf_iri or pp[0] == rdfs_iri or pp[0] == owl_iri:
            ptype = HasProperty.Ptype.system
        elif pp[0] == knora_api_iri:
            ptype = HasProperty.Ptype.knora
        else:
            ptype = HasProperty.Ptype.other
            ontology_id = context.iri_from_prefix(pp[0])
        property_id = p
        return ontology_id, property_id, ptype

    @staticmethod
    def _fromJsonObj_get_cardinality(jsonld_obj: dict[str, Any], owl_iri: str):
        cardinality: Cardinality
        if jsonld_obj.get(owl_iri + ":cardinality") is not None:
            cardinality = Cardinality.C_1
        elif jsonld_obj.get(owl_iri + ":maxCardinality") is not None:
            cardinality = Cardinality.C_0_1
        elif jsonld_obj.get(owl_iri + ":minCardinality") is not None:
            if jsonld_obj.get(owl_iri + ":minCardinality") == 0:
                cardinality = Cardinality.C_0_n
            elif jsonld_obj.get(owl_iri + ":minCardinality") == 1:
                cardinality = Cardinality.C_1_n
            else:
                raise BaseError("Problem with cardinality")
        else:
            raise BaseError("Problem with cardinality")
        return cardinality

    def create(self, last_modification_date: DateTimeStamp) -> tuple[DateTimeStamp, ResourceClass]:
        if self._ontology_id is None:
            raise BaseError("Ontology id required")
        if self._property_id is None:
            raise BaseError("Property id required")
        if self._cardinality is None:
            raise BaseError("Cardinality id required")

        jsonobj = self._toJsonObj_create(last_modification_date)
        result = self._con.post(HasProperty.ROUTE, jsonobj)
        last_modification_date = DateTimeStamp(result["knora-api:lastModificationDate"])
        return last_modification_date, ResourceClass.fromJsonObj(self._con, self._context, result["@graph"])

    def update(self, last_modification_date: DateTimeStamp) -> tuple[DateTimeStamp, ResourceClass]:
        if self._ontology_id is None:
            raise BaseError("Ontology id required")
        if self._property_id is None:
            raise BaseError("Property id required")
        if self._cardinality is None:
            raise BaseError("Cardinality id required")
        jsonobj = self._toJsonObj_update(last_modification_date)
        result = self._con.put(HasProperty.ROUTE, jsonobj)
        last_modification_date = DateTimeStamp(result["knora-api:lastModificationDate"])
        return last_modification_date, ResourceClass.fromJsonObj(self._con, self._context, result["@graph"])

    def _toJsonObj_update(self, lastModificationDate: DateTimeStamp) -> dict[str, Any]:
        occurrence = self._toJsonObj_get_owl_cardinality()
        tmp = {
            "@id": self._ontology_id,
            "@type": "owl:Ontology",
            "knora-api:lastModificationDate": lastModificationDate.toJsonObj(),
            "@graph": [
                {
                    "@id": self._resclass_id,
                    "@type": "owl:Class",
                    "rdfs:subClassOf": {
                        "@type": "owl:Restriction",
                        occurrence[0]: occurrence[1],
                        "owl:onProperty": {"@id": self._property_id},
                    },
                }
            ],
            "@context": self._context.toJsonObj(),
        }
        if self._gui_order is not None and "gui_order" in self._changed:
            tmp["@graph"][0]["rdfs:subClassOf"]["salsah-gui:guiOrder"] = self._gui_order
        return tmp

    def _toJsonObj_create(self, lastModificationDate: DateTimeStamp) -> dict[str, Any]:
        occurrence = self._toJsonObj_get_owl_cardinality()
        tmp = {
            "@id": self._ontology_id,
            "@type": "owl:Ontology",
            "knora-api:lastModificationDate": lastModificationDate.toJsonObj(),
            "@graph": [
                {
                    "@id": self._resclass_id,
                    "@type": "owl:Class",
                    "rdfs:subClassOf": {
                        "@type": "owl:Restriction",
                        occurrence[0]: occurrence[1],
                        "owl:onProperty": {"@id": self._property_id},
                    },
                }
            ],
            "@context": self._context.toJsonObj(),
        }
        if self._gui_order is not None:
            tmp["@graph"][0]["rdfs:subClassOf"]["salsah-gui:guiOrder"] = self._gui_order
        return tmp

    def _toJsonObj_get_owl_cardinality(self) -> tuple[str, int]:
        if self._cardinality is None:
            raise BaseError("There must be a cardinality given!")
        switcher = {
            Cardinality.C_1: ("owl:cardinality", 1),
            Cardinality.C_0_1: ("owl:maxCardinality", 1),
            Cardinality.C_0_n: ("owl:minCardinality", 0),
            Cardinality.C_1_n: ("owl:minCardinality", 1),
        }
        return switcher.get(self._cardinality)

    def createDefinitionFileObj(self, context: Context, shortname: str) -> dict[str, Any]:
        cardinality = {}
        if self._ptype == HasProperty.Ptype.other or self.property_id in [
            "knora-api:isPartOf",
            "knora-api:seqnum",
        ]:
            cardinality["propname"] = context.reduce_iri(self._property_id, shortname)
            cardinality["cardinality"] = self._cardinality.value
            if self._gui_order is not None:
                cardinality["gui_order"] = self._gui_order
        return cardinality


class ResourceClass(Model):
    """
    This class represents a DSP resource class

    Attributes
    ----------

    con : Connection
        A Connection instance to a DSP server

    iri : str
        IRI of the ResourceClass [readonly, cannot be modified after creation of instance]

    name: str
        The name of the resource class, e.g. "Book", "Person", "Portait". Usually these names start
        with a capital letter

    ontology_id: str
        The IRI/Id of the ontology this resource class belongs to

    superclasses: str, list[str]
        This is a list of superclasses for this resource class. Usually a project specific class must at least
        be a subclass of "Resource", but can be subclassed of any other valid resource class. In addition, external
        ontologies may be referenced:
        e.g.:
        ```
        "super": "Resource"

        "super": ":MySpecialTYpe"

        "super": [":MySpecialType", "dcterms:BibliographicResource"]
        ```

    label: language dependent string, that is a dict like {"en": "Biblography", "de": "Literaturverzeichnis"}
        A label (human readable name) for the resource

    comment: language dependent string, that is a dict like {"en": "a comment", "de": "Ein Kommentar"}
        A commentary to further explain what this resource class is used for

    permission: str
        The default permissions to be used if an instance of this resource class is being created

    has_properties: dict[str, HasProperty]
        Holds a dict with the property names as keys and a HasProperty instance. HasProperty holds
        the information, how this resource class uses this property (basically the cardinality)

    changed: bool
        is set to True, if one of the fields has been chaned by the user (internal use only!)


    Methods
    -------

    getProperty: Get information about a property defined for this resource class
        getPropertery(prop_id: str) -> HasProperty
        returns a HasProperty-instance

    addProperty: Add a new property to the resource class
        addProperty(property_id: str, cardinality: Cardinality, last_modification_date: DateTimeStamp)
        -> Optional[DateTimeStamp]

    create: Create a new resource class on the connected server

    update: Update the information of a resource class on the connected server

    delete: Mark a resource class as deleted (on the connected server)

    createDefinitionFileObj: Create an object suitable for jsonification that conforms the the DSP-TOOLS ontology tools

    print: Print the content of this class to the console

    """

    ROUTE: str = "/v2/ontologies/classes"

    _context: Context
    _iri: str
    _name: str
    _ontology_id: str
    _superclasses: list[str]
    _label: LangString
    _comment: LangString
    _permissions: str
    _has_properties: dict[str, HasProperty]

    def __init__(
        self,
        con: Connection,
        context: Context,
        iri: Optional[str] = None,
        name: Optional[str] = None,
        ontology_id: Optional[str] = None,
        superclasses: Optional[Sequence[Union[ResourceClass, str]]] = None,
        label: Optional[Union[LangString, str]] = None,
        comment: Optional[Union[LangString, str]] = None,
        permissions: Optional[str] = None,
        has_properties: Optional[dict[str, HasProperty]] = None,
    ):
        """
        Create an instance of  ResourceClass

        :param con:
        :param context:
        :param iri:
        :param name:
        :param ontology_id:
        :param superclasses:
        :param label:
        :param comment:
        :param permissions:
        :param has_properties:
        """
        super().__init__(con)
        self._context = context
        self._iri = iri
        self._name = name
        if ontology_id is not None:
            self._ontology_id = context.iri_from_prefix(ontology_id)
        if isinstance(superclasses, ResourceClass):
            self._superclasses = list(map(lambda a: a.iri, superclasses))
        else:
            self._superclasses = superclasses

        self._label = self._check_process_langstring(label, "label")
        self._comment = self._check_process_langstring(comment, "comment")
        self._permissions = permissions
        self._has_properties = has_properties
        self._changed = set()

    @staticmethod
    def _check_process_langstring(langstring_to_check: None | str | LangString, what: str) -> LangString:
        if langstring_to_check is not None:
            if isinstance(langstring_to_check, str):
                return LangString(langstring_to_check)
            elif isinstance(langstring_to_check, LangString):
                return langstring_to_check
            else:
                raise BaseError(f"Invalid LangString for {what}!")
        else:
            return LangString({})

    #
    # Here follows a list of getters/setters
    #
    @property
    def name(self) -> Optional[str]:
        return self._name

    @property
    def iri(self) -> Optional[str]:
        return self._iri

    @property
    def ontology_id(self) -> Optional[str]:
        return self._ontology_id

    @property
    def superclasses(self) -> Optional[list[str]]:
        return self._superclasses

    @property
    def label(self) -> LangString:
        return self._label

    @label.setter
    def label(self, value: Optional[Union[LangString, str]]) -> None:
        if value is None:
            self._label.empty()  # clear all labels
        elif isinstance(value, LangString):
            self._label = value
        elif isinstance(value, str):
            self._label = LangString(value)
        else:
            raise BaseError("Not a valid LangString")
        self._changed.add("label")

    @property
    def comment(self) -> LangString:
        return self._comment

    @comment.setter
    def comment(self, value: Optional[LangString]) -> None:
        if value is None:
            self._comment.empty()  # clear all comments!
        elif isinstance(value, LangString):
            self._comment = value
        elif isinstance(value, str):
            self._comment = LangString(value)
        else:
            raise BaseError("Not a valid LangString")
        self._changed.add("comment")

    @property
    def permissions(self) -> Optional[str]:
        return self._permissions

    @property
    def has_properties(self) -> dict[str, HasProperty]:
        return self._has_properties

    def getProperty(self, property_id: str) -> Optional[HasProperty]:
        if self._has_properties is None:
            return None
        else:
            return self._has_properties.get(self._context.get_prefixed_iri(property_id))

    def addProperty(
        self,
        last_modification_date: DateTimeStamp,
        property_id: str,
        cardinality: Cardinality,
        gui_order: Optional[int] = None,
    ) -> DateTimeStamp:
        if self._has_properties.get(property_id) is None:
            latest_modification_date, resclass = HasProperty(
                con=self._con,
                context=self._context,
                ontology_id=self._ontology_id,
                property_id=property_id,
                resclass_id=self.iri,
                cardinality=cardinality,
                gui_order=gui_order,
            ).create(last_modification_date)
            hp = resclass.getProperty(property_id)
            hp.ontology_id = self._context.iri_from_prefix(self._ontology_id)
            hp.resclass_id = self._iri
            self._has_properties[hp.property_id] = hp
            return latest_modification_date
        else:
            raise BaseError("Property already has cardinality in this class! " + property_id)

    @classmethod
    def fromJsonObj(cls, con: Connection, context: Context, json_obj: Any) -> ResourceClass:
        if isinstance(json_obj, list):
            json_obj = json_obj[0]
        rdfs = context.prefix_from_iri("http://www.w3.org/2000/01/rdf-schema#")
        owl = context.prefix_from_iri("http://www.w3.org/2002/07/owl#")
        knora_api = context.prefix_from_iri("http://api.knora.org/ontology/knora-api/v2#")

        if not (json_obj.get(knora_api + ":isResourceClass") or json_obj.get(knora_api + ":isStandoffClass")):
            raise BaseError("This is not a resource!")

        if json_obj.get("@id") is None:
            raise BaseError('Resource class has no "@id"!')

        tmp_id = json_obj.get("@id").split(":")
        iri = context.iri_from_prefix(tmp_id[0]) + "#" + tmp_id[1]
        ontology_id = tmp_id[0]
        name = tmp_id[1]

        has_properties, superclasses = cls._fromJsonObj_get_superclass(con, context, json_obj, rdfs, owl)

        label = LangString.fromJsonLdObj(json_obj.get(rdfs + ":label"))
        comment = LangString.fromJsonLdObj(json_obj.get(rdfs + ":comment"))
        return cls(
            con=con,
            context=context,
            name=name,
            iri=iri,
            ontology_id=ontology_id,
            superclasses=superclasses,
            label=label,
            comment=comment,
            has_properties=has_properties,
        )

    @staticmethod
    def _fromJsonObj_get_superclass(
        con: Connection, context: Context, json_obj: dict[str, Any], rdfs: str, owl: str
    ) -> tuple[None | dict[str, HasProperty], None | list[str]]:
        superclasses_obj = json_obj.get(rdfs + ":subClassOf")
        if superclasses_obj is not None:
            supercls = list(filter(lambda a: a.get("@id") is not None, superclasses_obj))
            superclasses = list(map(lambda a: a["@id"], supercls))
            has_props = list(filter(lambda a: a.get("@type") == (owl + ":Restriction"), superclasses_obj))
            has_properties = dict(map(lambda a: HasProperty.fromJsonObj(con, context, a), has_props))
            #
            # now remove the ...Value stuff from resource pointers: A resource pointer is returned as 2 properties:
            # one a direct link, the other the pointer to a link value
            #
            tmp = dict(has_properties)
            for key in tmp.keys():
                if key.endswith("Value"):
                    key_without_value = key.removesuffix("Value")
                    if key_without_value in has_properties:
                        del has_properties[key]
        else:
            superclasses = None
            has_properties = None
        return has_properties, superclasses

    def create(self, last_modification_date: DateTimeStamp) -> tuple[DateTimeStamp, ResourceClass]:
        jsonobj = self._toJsonObj_create(last_modification_date)
        result = self._con.post(ResourceClass.ROUTE, jsonobj)
        last_modification_date = DateTimeStamp(result["knora-api:lastModificationDate"])
        return last_modification_date, ResourceClass.fromJsonObj(self._con, self._context, result["@graph"])

    def update(self, last_modification_date: DateTimeStamp) -> tuple[DateTimeStamp, ResourceClass]:
        #
        # Note: DSP is able to change only one thing per call, either label or comment!
        #
        result = None
        something_changed = False
        if "label" in self._changed:
            jsonobj = self._toJsonObj_update(last_modification_date, "label")
            result = self._con.put(ResourceClass.ROUTE, jsonobj)
            last_modification_date = DateTimeStamp(result["knora-api:lastModificationDate"])
            something_changed = True
        if "comment" in self._changed:
            jsonobj = self._toJsonObj_update(last_modification_date, "comment")
            result = self._con.put(ResourceClass.ROUTE, jsonobj)
            last_modification_date = DateTimeStamp(result["knora-api:lastModificationDate"])
            something_changed = True
        if something_changed:
            return last_modification_date, ResourceClass.fromJsonObj(self._con, self._context, result["@graph"])
        else:
            return last_modification_date, self

    def _toJsonObj_update(self, lastModificationDate: DateTimeStamp, what_changed: str) -> dict[str, Any]:
        ontid, resid = self._get_onto_res_iri()
        tmp = {
            "@id": ontid,  # self._ontology_id,
            "@type": "owl:Ontology",
            "knora-api:lastModificationDate": lastModificationDate.toJsonObj(),
            "@graph": [
                {
                    "@id": resid,
                    "@type": "owl:Class",
                }
            ],
            "@context": self._context.toJsonObj(),
        }
        if what_changed == "label":
            if not self._label.isEmpty() and "label" in self._changed:
                tmp["@graph"][0]["rdfs:label"] = self._label.toJsonLdObj()
        if what_changed == "comment":
            if not self._comment.isEmpty() and "comment" in self._changed:
                tmp["@graph"][0]["rdfs:comment"] = self._comment.toJsonLdObj()
        return tmp

    def _toJsonObj_create(self, lastModificationDate: DateTimeStamp) -> dict[str, Any]:
        ontid, resid = self._get_onto_res_iri()
        if self._name is None:
            raise BaseError("There must be a valid resource class name!")
        if self._ontology_id is None:
            raise BaseError("There must be a valid ontology_id given!")
        if self._superclasses is None:
            superclasses = [{"@id": "knora-api:Resource"}]
        else:
            superclasses = list(map(self._resolve_resref, self._superclasses))
        if self._label is None or self._label.isEmpty():
            self._label = LangString("no label available")
        tmp = {
            "@id": ontid,  # self._ontology_id,
            "@type": "owl:Ontology",
            "knora-api:lastModificationDate": lastModificationDate.toJsonObj(),
            "@graph": [
                {
                    "@id": resid,
                    "@type": "owl:Class",
                    "rdfs:label": self._label.toJsonLdObj(),
                    "rdfs:subClassOf": superclasses,
                }
            ],
            "@context": self._context.toJsonObj(),
        }
        if self._comment:
            tmp["@graph"][0]["rdfs:comment"] = self._comment.toJsonLdObj()
        return tmp

    def _get_onto_res_iri(self) -> tuple[str, str]:
        exp = regex.compile("^http.*")  # It is already a fully IRI
        if exp.match(self._ontology_id):
            resid = self._context.prefix_from_iri(self._ontology_id) + ":" + self._name
            ontid = self._ontology_id
        else:
            resid = self._ontology_id + ":" + self._name
            ontid = self._context.iri_from_prefix(self._ontology_id)
        return ontid, resid

    def _resolve_resref(self, resref: str) -> dict[str, str]:
        tmp = resref.split(":")
        if len(tmp) > 1:
            if tmp[0] and self._context.iri_from_prefix(tmp[0]) != self._ontology_id:
                self._context.add_context(tmp[0])
                return {"@id": resref}  # fully qualified name in the form "prefix:name"
            else:
                return {
                    "@id": self._context.prefix_from_iri(self._ontology_id) + ":" + tmp[1]
                }  # ":name" in current ontology
        else:
            return {"@id": "knora-api:" + resref}  # no ":", must be from knora-api!

    def delete(self, last_modification_date: DateTimeStamp) -> DateTimeStamp:
        result = self._con.delete(
            ResourceClass.ROUTE + "/" + quote_plus(self._iri) + "?lastModificationDate=" + str(last_modification_date)
        )
        return DateTimeStamp(result["knora-api:lastModificationDate"])

    def createDefinitionFileObj(self, context: Context, shortname: str, skiplist: list[str]) -> dict[str, Any]:
        resource = {"name": self._name}
        if self._superclasses:
            resource["super"] = self._create_DefinitionFileObj_superclass(context, shortname)
        resource["labels"] = self._label.createDefinitionFileObj()
        if not self._comment.isEmpty():
            resource["comments"] = self._comment.createDefinitionFileObj()
        if self._has_properties:
            self._create_DefinitionFileObj_cardinalities(context, resource, shortname, skiplist)
        return resource

    def _create_DefinitionFileObj_cardinalities(
        self, context: Context, resource: dict[str, str], shortname: str, skiplist: list[str]
    ) -> dict[str, Any]:
        cardinalities: list[dict[str, Any]] = []
        for _, hp in self._has_properties.items():
            if hp.property_id in skiplist:
                continue
            if hp.ptype == HasProperty.Ptype.other or hp.property_id in [
                "knora-api:isPartOf",
                "knora-api:seqnum",
            ]:
                cardinalities.append(hp.createDefinitionFileObj(context, shortname))
        if cardinalities:
            resource["cardinalities"] = cardinalities
        return resource

    def _create_DefinitionFileObj_superclass(self, context: Context, shortname: str) -> list[str] | str:
        if len(self._superclasses) > 1:
            superclasses = []
            for sc in self._superclasses:
                superclasses.append(context.reduce_iri(sc, shortname))
        else:
            superclasses = context.reduce_iri(self._superclasses[0], shortname)
        return superclasses



import warnings

import pytest
from lxml import etree

from dsp_tools.models.custom_warnings import DspToolsUserWarning
from dsp_tools.xmllib.models.values import BooleanValue
from dsp_tools.xmllib.models.values import ColorValue
from dsp_tools.xmllib.models.values import DateValue
from dsp_tools.xmllib.models.values import DecimalValue
from dsp_tools.xmllib.models.values import GeonameValue
from dsp_tools.xmllib.models.values import IntValue
from dsp_tools.xmllib.models.values import LinkValue
from dsp_tools.xmllib.models.values import ListValue
from dsp_tools.xmllib.models.values import Richtext
from dsp_tools.xmllib.models.values import SimpleText
from dsp_tools.xmllib.models.values import TimeValue
from dsp_tools.xmllib.models.values import UriValue


class TestBooleanValue:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            BooleanValue("False", ":booleanProp", resource_id="res_id")
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            BooleanValue("other", ":booleanProp", resource_id="res_id")

    def test_serialise(self) -> None:
        v = BooleanValue("0", ":booleanProp", resource_id="res_id")
        expected = (
            b"<boolean-prop "
            b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            b'name=":booleanProp">'
            b'<boolean permissions="prop-default">false</boolean>'
            b"</boolean-prop>"
        )
        res_str = etree.tostring(v.serialise())
        assert res_str == expected


class TestColorValue:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            ColorValue("#FFFFFF", ":colorProp", resource_id="res_id")
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            ColorValue("invalidColor", ":colorProp", resource_id="res_id")

    def test_serialise(self) -> None:
        v = ColorValue("#000000", ":colorProp", resource_id="res_id")
        expected = (
            b"<color-prop "
            b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            b'name=":colorProp">'
            b'<color permissions="prop-default">#000000</color>'
            b"</color-prop>"
        )
        res_str = etree.tostring(v.serialise())
        assert res_str == expected


class TestDateValue:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            DateValue("2023-01-01", ":dateProp", resource_id="res_id")
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            DateValue("invalidDate", ":dateProp", resource_id="res_id")

    def test_serialise(self) -> None:
        v = DateValue("2023-01-01", ":dateProp", resource_id="res_id")
        expected = (
            b"<date-prop "
            b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            b'name=":dateProp">'
            b'<date permissions="prop-default">2023-01-01</date>'
            b"</date-prop>"
        )
        res_str = etree.tostring(v.serialise())
        assert res_str == expected


class TestDecimalValue:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            DecimalValue("3.14", ":decimalProp", resource_id="res_id")
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            DecimalValue("invalidDecimal", ":decimalProp", resource_id="res_id")

    def test_serialise(self) -> None:
        v = DecimalValue("3.14", ":decimalProp", resource_id="res_id")
        expected = (
            b"<decimal-prop "
            b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            b'name=":decimalProp">'
            b'<decimal permissions="prop-default">3.14</decimal>'
            b"</decimal-prop>"
        )
        res_str = etree.tostring(v.serialise())
        assert res_str == expected


class TestGeonameValue:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            GeonameValue("00099", ":geonameProp", resource_id="res_id")
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            GeonameValue("invalidGeoname", ":geonameProp", resource_id="res_id")

    def test_serialise(self) -> None:
        v = GeonameValue("99", ":geonameProp", resource_id="res_id")
        expected = (
            b"<geoname-prop "
            b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            b'name=":geonameProp">'
            b'<geoname permissions="prop-default">99</geoname>'
            b"</geoname-prop>"
        )
        res_str = etree.tostring(v.serialise())
        assert res_str == expected


class TestIntValue:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            IntValue("42", ":intProp", resource_id="res_id")
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            IntValue("invalidInt", ":intProp", resource_id="res_id")

    def test_serialise(self) -> None:
        v = IntValue("42", ":intProp", resource_id="res_id")
        expected = (
            b"<integer-prop "
            b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            b'name=":intProp">'
            b'<integer permissions="prop-default">42</integer>'
            b"</integer-prop>"
        )
        res_str = etree.tostring(v.serialise())
        assert res_str == expected


class TestLinkValue:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            LinkValue("link", ":linkProp", resource_id="res_id")
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            LinkValue(None, ":linkProp", resource_id="res_id")  # type: ignore[arg-type]

    def test_serialise(self) -> None:
        v = LinkValue("res_link", ":linkProp", resource_id="res_id")
        expected = (
            b"<resptr-prop "
            b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            b'name=":linkProp">'
            b'<resptr permissions="prop-default">res_link</resptr>'
            b"</resptr-prop>"
        )
        res_str = etree.tostring(v.serialise())
        assert res_str == expected


class TestListValue:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            ListValue("item1", "listName", ":listProp", resource_id="res_id")
        assert len(caught_warnings) == 0

    def test_warns_false_node(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            ListValue(None, "list", ":listProp", resource_id="res_id")

    def test_warns_false_list(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            ListValue("Node", None, ":listProp", resource_id="res_id")

    def test_serialise(self) -> None:
        v = ListValue("item1", "listName", ":listProp", resource_id="res_id")
        expected = (
            b"<list-prop "
            b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            b'name=":listProp" list="listName">'
            b'<list permissions="prop-default">item1</list>'
            b"</list-prop>"
        )
        res_str = etree.tostring(v.serialise())
        assert res_str == expected


class TestRichtext:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            Richtext("<p>Hello World</p>", ":richtextProp", resource_id="res_id")
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            Richtext(None, ":richtextProp", resource_id="res_id")  # type: ignore[arg-type]

    def test_serialise(self) -> None:
        v = Richtext("<p>Hello World</p>", ":richtextProp", resource_id="res_id")
        expected = (
            b"<text-prop "
            b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            b'name=":richtextProp">'
            b'<text encoding="xml" permissions="prop-default">&lt;p&gt;Hello World&lt;/p&gt;</text>'
            b"</text-prop>"
        )
        res_str = etree.tostring(v.serialise())
        assert res_str == expected


class TestSimpleText:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            SimpleText("Hello World", ":simpleTextProp", resource_id="res_id")
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            SimpleText(None, ":simpleTextProp", resource_id="res_id")  # type: ignore[arg-type]

    def test_serialise(self) -> None:
        v = SimpleText("Hello World", ":simpleTextProp", resource_id="res_id")
        expected = (
            b"<text-prop "
            b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            b'name=":simpleTextProp">'
            b'<text encoding="utf8" permissions="prop-default">Hello World</text>'
            b"</text-prop>"
        )
        res_str = etree.tostring(v.serialise())
        assert res_str == expected


class TestTimeValue:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            TimeValue("2009-10-10T12:00:00-05:00", ":timeProp", resource_id="res_id")
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            TimeValue("invalidTime", ":timeProp", resource_id="res_id")

    def test_serialise(self) -> None:
        v = TimeValue("2009-10-10T12:00:00-05:00", ":timeProp", resource_id="res_id")
        expected = (
            b"<time-prop "
            b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            b'name=":timeProp">'
            b'<time permissions="prop-default">2009-10-10T12:00:00-05:00</time>'
            b"</time-prop>"
        )
        res_str = etree.tostring(v.serialise())
        assert res_str == expected


class TestUriValue:
    def test_good(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            UriValue("https://example.com", ":uriProp", resource_id="res_id")
        assert len(caught_warnings) == 0

    def test_warns(self) -> None:
        with pytest.warns(DspToolsUserWarning):
            UriValue("invalidUri", ":uriProp", resource_id="res_id")

    def test_serialise(self) -> None:
        v = UriValue("https://example.com", ":uriProp", resource_id="res_id")
        expected = (
            b"<uri-prop "
            b'xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            b'name=":uriProp">'
            b'<uri permissions="prop-default">https://example.com</uri>'
            b"</uri-prop>"
        )
        res_str = etree.tostring(v.serialise())
        assert res_str == expected


if __name__ == "__main__":
    pytest.main([__file__])



from dataclasses import dataclass
from typing import Any
from typing import cast

import requests
from loguru import logger
from rdflib import Graph
from requests import ReadTimeout
from requests import RequestException
from requests import Response

from dsp_tools.models.exceptions import InternalError
from dsp_tools.models.exceptions import UserError


@dataclass
class OntologyConnection:
    api_url: str
    shortcode: str

    def _get(self, url: str, headers: dict[str, Any] | None = None) -> Response:
        """
        Sends a get request to the designated url

        Args:
            url: URL for the request
            headers: headers for the request

        Returns:
            Response of the request if it was a 200 response code

        Raises:
            InternalError: in case of errors raised by the requests library
            UserError: If a non-200 response was given
        """
        try:
            timeout = 100
            logger.debug(f"REQUEST: GET to {url}, timeout: {timeout}, headers: {headers}")
            response = requests.get(url, headers=headers, timeout=timeout)
            logger.debug(f"RESPONSE: {response.status_code}")
        except (TimeoutError, ReadTimeout) as err:
            logger.exception(err)
            raise InternalError("TimeoutError occurred. See logs for details.") from None
        except (ConnectionError, RequestException) as err:
            logger.exception(err)
            raise InternalError("ConnectionError occurred. See logs for details.") from None
        if not response.ok:
            msg = f"Non-ok response: {response.status_code}\nOriginal message: {response.text}"
            logger.exception(msg)
            raise UserError(msg)
        return response

    def get_ontologies(self) -> list[str]:
        """
        Returns a list of project ontologies as a string in turtle format.

        Returns:
            list of ontologies
        """
        ontology_iris = self._get_ontology_iris()
        return [self._get_one_ontology(x) for x in ontology_iris]

    def _get_ontology_iris(self) -> list[str]:
        endpoint = f"{self.api_url}/admin/projects/shortcode/{self.shortcode}"
        response = self._get(endpoint)
        response_json = cast(dict[str, Any], response.json())
        msg = f"The response from the API does not contain any ontologies.\nAPI response:{response.text}"
        if not (proj := response_json.get("project")):
            logger.exception(msg)
            raise UserError(msg)
        if not (ontos := proj.get("ontologies")):
            logger.exception(msg)
            raise UserError(msg)
        output = cast(list[str], ontos)
        return output

    def _get_one_ontology(self, ontology_iri: str) -> str:
        response = self._get(ontology_iri, headers={"Accept": "text/turtle"})
        return response.text


@dataclass
class ShaclValidator:
    """Client to validate RDF data against a given SHACL shape."""

    dsp_api_url: str

    def validate(self, data_ttl: str, shacl_ttl: str) -> Graph:
        """
        Sends a multipart/form-data request with two turtle files (data.ttl and shacl.ttl) to the given URL
        and expects a response containing a single text/turtle body which is loaded into an rdflib Graph.

        Args:
            data_ttl (str): The turtle content for the data.ttl file (as a string).
            shacl_ttl (str): The turtle content for the shacl.ttl file (as a string).

        Returns:
            Graph: The rdflib Graph object loaded with the response turtle data.

        Raises:
            InternalError: in case of a non-ok response
        """
        files = {
            "data.ttl": ("data.ttl", data_ttl, "text/turtle"),
            "shacl.ttl": ("shacl.ttl", shacl_ttl, "text/turtle"),
        }
        timeout = 10
        request_url = f"{self.dsp_api_url}/shacl/validate"
        logger.debug(f"REQUEST: POST to {request_url}, timeout: {timeout}")
        response = requests.post(request_url, files=files, timeout=timeout)
        if not response.ok:
            msg = f"Failed to send request. Status code: {response.status_code}, Original Message:\n{response.text}"
            logger.error(msg)
            raise InternalError(msg)
        graph = Graph()
        graph.parse(data=response.text, format="turtle")
        return graph



from typing import Any

import pandas as pd
import pytest

from dsp_tools.xmllib.value_converters import convert_to_bool_string


@pytest.mark.parametrize("val", ["false", "0  ", "  0.0", "NO", False])
def test_convert_to_bool_false(val: Any) -> None:
    assert convert_to_bool_string(val) == "false"


@pytest.mark.parametrize("val", ["TRUE ", "  1", "1.0", "Yes", True])
def test_convert_to_bool_true(val: Any) -> None:
    assert convert_to_bool_string(val) == "true"


@pytest.mark.parametrize("val", [pd.NA, None, 2.421, 10, "other", "  "])
def test_convert_to_bool_failure(val: Any) -> None:
    assert convert_to_bool_string(val) == str(val)


if __name__ == "__main__":
    pytest.main([__file__])



import pytest

from dsp_tools.utils.uri_util import is_iiif_uri


@pytest.mark.parametrize(
    "uri",
    [
        "https://iiif.dasch.swiss/0811/1Oi7mdiLsG7-FmFgp0xz2xU.jp2/full/837,530/0/default.jp2",
        "https://iiif.dasch.swiss/0811/1Oi7mdiLsG7-FmFgp0xz2xU.jp2/2048,0,1298,2048/649,1024/0/default.jpg",
        "https://iiif.io/api/image/3.0/example/reference/1-newspaper-p2/full/max/0/default.webp",
        "https://newspapers.library.wales/iiif/2.0/image/4497470/512,2048,512,512/256,/0/default.jpg",
        "https://example.org/image-service/abcd1234/full/max/0/default.jpg",
        "https://www.example.org/prefix1/abcd1234/80,15,60,75/full/0/native",
        "http://www.example.org/prefix1/abcd1234/80,15,60.6,75/full/0/native.jpg",
        "http://www.example.org/prefix1/abcd1234/pct:10.5,10,80,70/full/0/native.jpg",
        "http://www.example.org/prefix1/abcd1234/full/full/0/native.jpg",
        "http://www.example.org/prefix1/abcd1234/full/600,/0/colour.jpg",
        "http://www.example.org/prefix1/abcd1234/full/600,/!1/color.jpg",
        "http://www.example.org/prefix1/abcd1234/full/600,/0/grey.jpg",
        "http://www.example.org/prefix1/abcd1234/full/600,/0/bitonal.JPG",
        "http://www.example.org/prefix1/prefix2/prefix3/prefix4/abcd1234/full/full/0/default.jpg",
        "https://www.example.org/prefix1/prefix2/prefix3/prefix4/abcd1234/full/full/!90/gray.webp",
        "http://www.example.org/prefix1/prefix2/prefix3/prefix4/abcd1234/full/max/0/default.jpg",
        "http://www.example.org/prefix1/abcd1234/80,15,60,75/full/0/native",
        "https://www.example.org/prefix1/prefix2/prefix3/prefix4/abcd1234/full/%5Emax/0/gray.webp",
        "https://iiif.dasch.swiss/0811/1Oi7mdiLsG7-FmFgp0xz2xU.jp2/full/max/0/default.jpg",
        "http://www.example.org/prefix1/prefix2/prefix3/abcd1234/80,15,60,75/full/0/native",
    ],
)
def test_is_iiif_uri_correct(uri: str) -> None:
    assert is_iiif_uri(uri)


@pytest.mark.parametrize(
    "uri",
    [
        "http://example.org",
        "https://iiif.dasch.swiss/0811/1Oi7mdiLsG7-FmFgp0xz2xU.jp2/2048,0,1298,2048/649,1024/0/default.jpg3",
        "https://iiif.ub.unibe.ch/image/v2.1/632664f2-20cb-43e4-8584-2fa3988c63a2/info.json",
        "https://iiif.dasch.swiss/0811/5Jd909CLmCJ-BUUL1DDOXGJ.jp2/info.json",
        "ftp://www.example.org/prefix1/prefix2/prefix3/prefix4/abcd1234/kljg/%5Emax/0/gray.webp",
        "http://www.example.org/prefix1/abcd1234/80.,15,60,75/full/0/native",
        "http://www.example.org/prefix1/abcd1234/.80,15,60,75/full/0/native",
        "http://www.example.org/prefix1/abcd1234/full/0/native.jpg",
        "http://www.example.org/prefix1/abcd1234/square/-----full/!90/gray.webp",
        "http://www.example.org/prefix1/abcd1234/square/---full/!90/gray.webp",
        "http://www.example.org/prefix1/abcd1234/square/full/---!90/gray.webp",
        "http://www.example.org/prefix1/abcd1234/square/full/!90/---gray.webp",
        "http://www.example.org/prefix1/abcd1234/-----square/full/!90/gray.webp",
    ],
)
def test_is_iiif_uri_wrong(uri: str) -> None:
    assert not is_iiif_uri(uri)


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from dataclasses import dataclass

import regex

from dsp_tools.commands.xmlupload.iri_resolver import IriResolver
from dsp_tools.models.exceptions import BaseError


@dataclass
class FormattedTextValue:
    """Represents a formatted text value with standard standoff markup"""

    xmlstr: str

    def as_xml(self) -> str:
        """Returns the formatted text value as XML (with XML declaration and wrapped in a <text> element)"""
        return f'<?xml version="1.0" encoding="UTF-8"?>\n<text>{self.xmlstr}</text>'

    def find_internal_ids(self) -> set[str]:
        """Returns a set of all internal ids found in the text value"""
        return set(regex.findall(pattern='href="IRI:(.*?):IRI"', string=self.xmlstr))

    def with_iris(self, iri_resolver: IriResolver) -> FormattedTextValue:
        """
        Returns a copy of this object, where all internal ids are replaced with iris according to the provided mapping.
        """
        s = self.xmlstr
        for internal_id in self.find_internal_ids():
            if iri := iri_resolver.get(internal_id):
                s = s.replace(f'href="IRI:{internal_id}:IRI"', f'href="{iri}"')
            else:
                raise BaseError(f"Internal ID {internal_id} could not be resolved to an IRI")
        return FormattedTextValue(s)



from datetime import datetime
from pathlib import Path
from time import sleep
from typing import cast

from loguru import logger
from tqdm import tqdm

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.commands.ingest_xmlupload.bulk_ingest_client import BulkIngestClient
from dsp_tools.utils.connection import Connection
from dsp_tools.utils.connection_live import ConnectionLive


def ingest_files(creds: ServerCredentials, shortcode: str) -> bool:
    """
    Kick off the ingest process on the server, and wait until it has finished.
    Then, retrieve the mapping CSV from the server and save it in the CWD.

    Args:
        creds: credentials to log in on the server
        shortcode: shortcode of the project

    Returns:
        success status
    """
    con: Connection = ConnectionLive(creds.server)
    con.login(creds.user, creds.password)
    bulk_ingest_client = BulkIngestClient(creds.dsp_ingest_url, con.get_token(), shortcode)
    bulk_ingest_client.trigger_ingest_process()
    sleep(5)
    mapping = _retrieve_mapping(bulk_ingest_client)
    _save_mapping(mapping, shortcode)
    return True


def _retrieve_mapping(bulk_ingest_client: BulkIngestClient) -> str:
    sleeping_time = 10
    desc = f"Wait until mapping CSV is ready. Ask server every {sleeping_time} seconds "
    progress_bar = tqdm(
        bulk_ingest_client.retrieve_mapping_generator(), desc=desc, bar_format="{desc}{elapsed}", dynamic_ncols=True
    )
    num_of_attempts = 0
    num_of_server_errors = 0
    for result in progress_bar:
        if result is False:
            num_of_attempts += 1
            num_of_server_errors += 1
        elif result is True:
            num_of_attempts += 1
        elif isinstance(result, str):
            break
        progress_bar.set_description(f"{desc}(attempts: {num_of_attempts}, server errors: {num_of_server_errors})")
        sleep(sleeping_time)
    return cast(str, result)


def _save_mapping(mapping: str, shortcode: str) -> None:
    filepath = Path(f"mapping-{shortcode}.csv")
    if filepath.exists():
        timestamp = datetime.now().strftime("%Y-%m-%d_%H.%M.%S")
        filepath = filepath.with_name(f"{filepath.stem}-{timestamp}.csv")
    filepath.write_text(mapping)
    print(f"Saved mapping CSV to '{filepath}'")
    logger.info(f"Saved mapping CSV to '{filepath}'")



from pathlib import Path

from lxml import etree
from rdflib import SH
from rdflib import Graph

from dsp_tools.commands.xml_validate.api_connection import OntologyConnection
from dsp_tools.commands.xml_validate.api_connection import ShaclValidator
from dsp_tools.commands.xml_validate.deserialise_input import deserialise_xml
from dsp_tools.commands.xml_validate.make_data_rdf import make_data_rdf
from dsp_tools.commands.xml_validate.models.data_deserialised import ProjectDeserialised
from dsp_tools.commands.xml_validate.models.data_rdf import DataRDF
from dsp_tools.commands.xml_validate.sparql.construct_shapes import construct_shapes_graph
from dsp_tools.utils.xml_utils import parse_xml_file
from dsp_tools.utils.xml_utils import remove_comments_from_element_tree
from dsp_tools.utils.xml_utils import transform_into_localnames
from dsp_tools.utils.xml_validation import validate_xml

LIST_SEPARATOR = "\n    - "


def xml_validate(filepath: Path, shortcode: str, api_url: str) -> None:
    """
    Takes a file and project information and validates it against the ontologies on the server.

    Args:
        filepath: path to the xml data file
        shortcode: shortcode of the project
        api_url: url of the api host
    """
    onto_con = OntologyConnection(api_url, shortcode)
    data_rdf = _get_data_info_from_file(filepath, api_url)
    ontologies = _get_project_ontos(onto_con)
    data_graph = data_rdf.make_graph() + ontologies
    val = ShaclValidator(api_url)
    conforms, result = _validate(val, ontologies, data_graph)
    if conforms:
        print("Validation passed!")
    else:
        print("Validation errors found!")


def _validate(validator: ShaclValidator, onto_graph: Graph, data_graph: Graph) -> tuple[bool, Graph]:
    what_is_validated = ["The following information of your data is being validated:", "Cardinalities"]
    print(LIST_SEPARATOR.join(what_is_validated))
    shapes = construct_shapes_graph(onto_graph)
    shape_str = shapes.serialize(format="ttl")
    data_str = data_graph.serialize(format="ttl")
    results = validator.validate(data_str, shape_str)
    conforms = bool(next(results.objects(None, SH.conforms)))
    return conforms, results


def _get_project_ontos(onto_con: OntologyConnection) -> Graph:
    all_ontos = onto_con.get_ontologies()
    g = Graph()
    for onto in all_ontos:
        og = Graph()
        og.parse(data=onto, format="ttl")
        g += og
    return g


def _get_data_info_from_file(file: Path, api_url: str) -> DataRDF:
    cleaned_root = _parse_and_clean_file(file, api_url)
    deserialised: ProjectDeserialised = deserialise_xml(cleaned_root)
    rdf_data: DataRDF = make_data_rdf(deserialised.data)
    return rdf_data


def _parse_and_clean_file(file: Path, api_url: str) -> etree._Element:
    root = parse_xml_file(file)
    root = remove_comments_from_element_tree(root)
    validate_xml(root)
    root = transform_into_localnames(root)
    return _replace_namespaces(root, api_url)


def _replace_namespaces(root: etree._Element, api_url: str) -> etree._Element:
    with open("src/dsp_tools/resources/xml_validate/replace_namespace.xslt", "rb") as xslt_file:
        xslt_data = xslt_file.read()
    shortcode = root.attrib["shortcode"]
    default_ontology = root.attrib["default-ontology"]
    namespace = f"{api_url}/ontology/{shortcode}/{default_ontology}/v2#"
    xslt_root = etree.XML(xslt_data)
    transform = etree.XSLT(xslt_root)
    replacement_value = etree.XSLT.strparam(namespace)
    return transform(root, replacementValue=replacement_value).getroot()



from datetime import datetime
from pathlib import Path

from viztracer import VizTracer

from dsp_tools.commands.xmlupload.stash.construct_and_analyze_graph import create_info_from_xml_for_graph
from dsp_tools.commands.xmlupload.stash.construct_and_analyze_graph import generate_upload_order
from dsp_tools.commands.xmlupload.stash.construct_and_analyze_graph import make_graph
from dsp_tools.utils.xml_utils import parse_and_clean_xml_file


def analyse_circles_in_data(xml_filepath: Path, tracer_output_file: str, save_tracer: bool = False) -> None:
    """
    This function takes an XML filepath
    It analyzes how many and which links have to be removed
    so that all circular references are broken up.
    This is for analysis purposes only.

    Args:
        xml_filepath: path to the file
        tracer_output_file: name of the file where the viztracer results should be saved
        save_tracer: True if the output of the viztracer should be saved
    """
    root = parse_and_clean_xml_file(xml_filepath)
    resptr_links, xml_links, all_resource_ids = create_info_from_xml_for_graph(root)
    tracer: VizTracer | None = None
    if save_tracer:
        tracer = VizTracer(
            minimize_memory=True,
            max_stack_depth=3,
        )
        tracer.start()
    start = datetime.now()
    print("=" * 80)
    print(f"Total Number of Resources: {len(all_resource_ids)}")
    print(f"Total Number of resptr Links: {len(resptr_links)}")
    print(f"Total Number of XML Texts with Links: {len(xml_links)}")
    print("=" * 80)
    graph, node_to_id, edges = make_graph(resptr_links, xml_links, all_resource_ids)
    _, _, stash_counter = generate_upload_order(graph, node_to_id, edges)
    end = datetime.now()
    if save_tracer and tracer:
        tracer.stop()
        tracer.save(output_file=tracer_output_file)
    print("Number of Links Stashed:", stash_counter)
    print("=" * 80)
    print("Start time:", start)
    print("End time:", end)
    print("Duration:", end - start)
    print("=" * 80)


if __name__ == "__main__":
    analyse_circles_in_data(
        xml_filepath=Path("testdata/xml-data/circular-references/test_circular_references_1.xml"),
        tracer_output_file="circular_references_tracer.json",
        save_tracer=False,
    )



import urllib.parse
from pathlib import Path
from typing import Any
from typing import Iterator

import pytest
import requests

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.commands.project.create.project_create import create_project
from dsp_tools.commands.xmlupload.xmlupload import xmlupload
from test.e2e.setup_testcontainers import get_containers

PROJECT_SHORTCODE = "4125"
ONTO_NAME = "e2e-testonto"
CREDS = ServerCredentials("root@example.com", "test", "http://0.0.0.0:3333")
ONTO_IRI = f"{CREDS.server}/ontology/{PROJECT_SHORTCODE}/{ONTO_NAME}/v2"
PROPS_IN_ONTO_JSON = 1
RESCLASSES_IN_ONTO_JSON = 2


@pytest.fixture
def _create_project() -> Iterator[None]:
    with get_containers():
        success = create_project(Path("testdata/json-project/test-project-e2e.json"), CREDS, verbose=True)
        assert success
        yield


@pytest.fixture
def _xmlupload(_create_project: None) -> None:
    success = xmlupload(Path("testdata/xml-data/test-data-e2e.xml"), CREDS, ".")
    assert success


@pytest.fixture
def auth_header(_create_project: None) -> dict[str, str]:
    payload = {"email": CREDS.user, "password": CREDS.password}
    token: str = requests.post(f"{CREDS.server}/v2/authentication", json=payload, timeout=3).json()["token"]
    return {"Authorization": f"Bearer {token}"}


@pytest.fixture
def project_iri(_create_project: None) -> str:
    get_project_route = f"{CREDS.server}/admin/projects/shortcode/{PROJECT_SHORTCODE}"
    project_iri: str = requests.get(get_project_route, timeout=3).json()["project"]["id"]
    return project_iri


@pytest.mark.usefixtures("_create_project")
def test_project(auth_header: dict[str, str]) -> None:
    get_project_url = f"{CREDS.server}/admin/projects/shortcode/{PROJECT_SHORTCODE}"
    project = requests.get(get_project_url, headers=auth_header, timeout=3).json()["project"]
    _check_project(project)

    onto = requests.get(ONTO_IRI, headers=auth_header, timeout=3).json()["@graph"]
    _check_props([elem for elem in onto if elem.get("knora-api:isResourceProperty")])
    _check_resclasses([elem for elem in onto if elem.get("knora-api:isResourceClass")])


@pytest.mark.usefixtures("_xmlupload")
def test_xmlupload(auth_header: dict[str, str], project_iri: str) -> None:
    img_resources = _get_resources(f"{ONTO_IRI}#ImageResource", auth_header, project_iri)
    _analyze_img_resources(img_resources)
    pdf_resources = _get_resources(f"{ONTO_IRI}#PDFResource", auth_header, project_iri)
    _analyze_pdf_resources(pdf_resources)


def _check_project(project: dict[str, Any]) -> None:
    assert project["shortname"] == "e2e-tp"
    assert project["shortcode"] == PROJECT_SHORTCODE
    assert project["longname"] == "e2e test project"
    assert project["description"] == [{"value": "The e2e test project", "language": "en"}]
    assert project["keywords"] == ["e2e-test-projec-keyword"]
    assert project["ontologies"] == [ONTO_IRI]


def _check_props(props: list[dict[str, Any]]) -> None:
    assert len(props) == PROPS_IN_ONTO_JSON
    assert props[0]["@id"] == f"{ONTO_NAME}:hasText"
    assert props[0]["rdfs:subPropertyOf"] == {"@id": "knora-api:hasValue"}
    assert props[0]["knora-api:objectType"] == {"@id": "knora-api:TextValue"}
    assert props[0]["rdfs:label"] == "Text"


def _check_resclasses(resclasses: list[dict[str, Any]]) -> None:
    assert len(resclasses) == RESCLASSES_IN_ONTO_JSON
    res_1, res_2 = resclasses

    assert res_1["@id"] == f"{ONTO_NAME}:ImageResource"
    assert res_1["rdfs:label"] == "Image Resource"
    cards_1 = res_1["rdfs:subClassOf"]
    fileval_card = [x for x in cards_1 if x.get("owl:onProperty", {}).get("@id") == "knora-api:hasStillImageFileValue"]
    assert len(fileval_card) == 1
    assert fileval_card[0].get("owl:cardinality") == 1
    hasText_card = [x for x in cards_1 if x.get("owl:onProperty", {}).get("@id") == f"{ONTO_NAME}:hasText"]
    assert len(hasText_card) == 1

    assert res_2["@id"] == f"{ONTO_NAME}:PDFResource"
    assert res_2["rdfs:label"] == "PDF Resource"


def _get_resources(resclass_iri: str, auth_header: dict[str, str], project_iri: str) -> list[dict[str, Any]]:
    resclass_iri_encoded = urllib.parse.quote_plus(resclass_iri)
    get_resources_route = f"{CREDS.server}/v2/resources?resourceClass={resclass_iri_encoded}&page=0"
    headers = auth_header | {"X-Knora-Accept-Project": project_iri}
    response = requests.get(get_resources_route, timeout=3, headers=headers).json()
    resources: list[dict[str, Any]] = response.get("@graph", [response])
    return resources


def _analyze_img_resources(img_resources: list[dict[str, Any]]) -> None:
    res_labels = sorted([res["rdfs:label"] for res in img_resources])
    assert res_labels == ["Resource 1", "Resource 2"]

    res_1 = next(res for res in img_resources if res["rdfs:label"] == "Resource 1")
    assert res_1.get("knora-api:hasStillImageFileValue")

    res_2 = next(res for res in img_resources if res["rdfs:label"] == "Resource 2")
    assert res_2.get("knora-api:hasStillImageFileValue")
    res_2_text_vals = sorted([x["knora-api:valueAsString"] for x in res_2[f"{ONTO_NAME}:hasText"]])
    assert res_2_text_vals == ["first text value", "second text value"]


def _analyze_pdf_resources(pdf_resources: list[dict[str, Any]]) -> None:
    res_labels = sorted([res["rdfs:label"] for res in pdf_resources])
    assert res_labels == ["Resource 3"]
    res_1 = next(res for res in pdf_resources if res["rdfs:label"] == "Resource 3")
    assert res_1.get("knora-api:hasDocumentFileValue")



import pandas as pd
import pytest

from dsp_tools.commands.excel2json.new_lists.models.deserialise import ColumnNodes
from dsp_tools.commands.excel2json.new_lists.models.deserialise import Columns
from dsp_tools.commands.excel2json.new_lists.models.deserialise import ExcelSheet


@pytest.fixture
def cols_en_list_only() -> Columns:
    return Columns(preferred_lang="en", list_cols=["en_list"], comment_cols=[], node_cols=[])


@pytest.fixture
def cols_en_1() -> Columns:
    n_1 = ColumnNodes(level_num=1, columns=["en_1"])
    return Columns(preferred_lang="en", list_cols=["en_list"], comment_cols=["en_comment"], node_cols=[n_1])


@pytest.fixture
def cols_en_1_no_comments() -> Columns:
    n_1 = ColumnNodes(level_num=1, columns=["en_1"])
    return Columns(preferred_lang="en", list_cols=["en_list"], comment_cols=[], node_cols=[n_1])


@pytest.fixture
def cols_en_de_1() -> Columns:
    n_1 = ColumnNodes(level_num=1, columns=["en_1", "de_1"])
    return Columns(
        preferred_lang="en",
        list_cols=["en_list", "de_list"],
        comment_cols=["en_comment", "de_comment"],
        node_cols=[n_1],
    )


@pytest.fixture
def cols_en_1_2() -> Columns:
    return Columns(
        preferred_lang="en",
        list_cols=["en_list"],
        comment_cols=["en_comments"],
        node_cols=[ColumnNodes(1, ["en_1"]), ColumnNodes(2, ["en_2"])],
    )


@pytest.fixture
def cols_en_de_1_2_no_comment() -> Columns:
    n_1 = ColumnNodes(level_num=1, columns=["en_1", "de_1"])
    n_2 = ColumnNodes(level_num=2, columns=["en_2", "de_2"])
    return Columns(preferred_lang="en", list_cols=["en_list", "de_list"], comment_cols=[], node_cols=[n_1, n_2])


@pytest.fixture
def cols_en_de_1_3() -> Columns:
    n_1 = ColumnNodes(level_num=1, columns=["en_1", "de_1"])
    n_2 = ColumnNodes(level_num=2, columns=["en_2", "de_2"])
    n_3 = ColumnNodes(level_num=3, columns=["en_3", "de_3"])
    return Columns(
        preferred_lang="en",
        list_cols=["en_list", "de_list"],
        comment_cols=["en_comments", "de_comments"],
        node_cols=[n_1, n_2, n_3],
    )


@pytest.fixture
def cols_en_de_1_3_no_comments() -> Columns:
    n_1 = ColumnNodes(level_num=1, columns=["en_1", "de_1"])
    n_2 = ColumnNodes(level_num=2, columns=["en_2", "de_2"])
    n_3 = ColumnNodes(level_num=3, columns=["en_3", "de_3"])
    return Columns(
        preferred_lang="en",
        list_cols=["en_list", "de_list"],
        comment_cols=[],
        node_cols=[n_1, n_2, n_3],
    )


@pytest.fixture
def f1_s1_good_en(cols_en_1_no_comments: Columns) -> ExcelSheet:
    df = pd.DataFrame(
        {
            "id (optional)": [11, 22, pd.NA, 44],
            "en_list": ["list1", "list1", "list1", "list1"],
            "en_1": [pd.NA, "node1", "node2", "node3"],
        }
    )
    return ExcelSheet(excel_name="file1", sheet_name="sheet1", df=df, col_info=cols_en_1_no_comments)


@pytest.fixture
def f2_s2_good_en_de(cols_en_de_1_3: Columns) -> ExcelSheet:
    df = pd.DataFrame(
        {
            "id (optional)": [1, 2, pd.NA, 4, 5, 6, 7, pd.NA],
            "en_list": ["list2", "list2", "list2", "list2", "list2", "list2", "list2", "list2"],
            "de_list": ["list2", "list2", "list2", "list2", "list2", "list2", "list2", "list2"],
            "en_comments": ["comment en list", pd.NA, pd.NA, "comment en node1.1.1", pd.NA, pd.NA, pd.NA, pd.NA],
            "de_comments": ["comment de list", pd.NA, pd.NA, "comment de node1.1.1", pd.NA, pd.NA, pd.NA, pd.NA],
            "en_1": [pd.NA, "node1", "node1", "node1", "node1", "node2", "node2", "node3"],
            "de_1": [pd.NA, "node1", "node1", "node1", "node1", "node2", "node2", "node3"],
            "en_2": [pd.NA, pd.NA, "node1.1", "node1.1", "node1.2", pd.NA, "node2.1", pd.NA],
            "de_2": [pd.NA, pd.NA, "node1.1", "node1.1", "node1.2", pd.NA, "node2.1", pd.NA],
            "en_3": [pd.NA, pd.NA, pd.NA, "node1.1.1", pd.NA, pd.NA, pd.NA, pd.NA],
            "de_3": [pd.NA, pd.NA, pd.NA, "node1.1.1", pd.NA, pd.NA, pd.NA, pd.NA],
        }
    )
    return ExcelSheet(excel_name="file2", sheet_name="sheet2", df=df, col_info=cols_en_de_1_3)


@pytest.fixture
def f1_s1_good_id_filled_empty_comments(cols_en_de_1_3_no_comments: Columns) -> ExcelSheet:
    df = pd.DataFrame(
        {
            "id": ["list_id", "1", "1.1", "2", "3", "3.1", "3.2", "3.2.1", "3.2.2"],
            "parent_id": ["list_id", "list_id", "1", "list_id", "list_id", "3", "3", "3.2", "3.2"],
            "en_comments": [pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA],
            "de_comments": [pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA],
            "en_list": [
                "Listname_en",
                "Listname_en",
                "Listname_en",
                "Listname_en",
                "Listname_en",
                "Listname_en",
                "Listname_en",
                "Listname_en",
                "Listname_en",
            ],
            "de_list": [
                "Listname_de",
                "Listname_de",
                "Listname_de",
                "Listname_de",
                "Listname_de",
                "Listname_de",
                "Listname_de",
                "Listname_de",
                "Listname_de",
            ],
            "en_1": [
                pd.NA,
                "Node_en_1",
                "Node_en_1",
                "Node_en_2",
                "Node_en_3",
                "Node_en_3",
                "Node_en_3",
                "Node_en_3",
                "Node_en_3",
            ],
            "de_1": [
                pd.NA,
                "Node_de_1",
                "Node_de_1",
                "Node_de_2",
                "Node_de_3",
                "Node_de_3",
                "Node_de_3",
                "Node_de_3",
                "Node_de_3",
            ],
            "en_2": [
                pd.NA,
                pd.NA,
                "Node_en_1.1",
                pd.NA,
                pd.NA,
                "Node_en_3.1",
                "Node_en_3.2",
                "Node_en_3.2",
                "Node_en_3.2",
            ],
            "de_2": [
                pd.NA,
                pd.NA,
                "Node_de_1.1",
                pd.NA,
                pd.NA,
                "Node_de_3.1",
                "Node_de_3.2",
                "Node_de_3.2",
                "Node_de_3.2",
            ],
            "en_3": [pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, "Node_en_3.2.1", "Node_en_3.2.2"],
            "de_3": [pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, "Node_de_3.2.1", "Node_de_3.2.2"],
        }
    )
    return ExcelSheet(excel_name="excel1", sheet_name="sheet1", df=df, col_info=cols_en_de_1_3_no_comments)


@pytest.fixture
def f1_s1_empty_node_column(cols_en_1_2: Columns) -> ExcelSheet:
    df = pd.DataFrame(
        {
            "id (optional)": [11, 22, pd.NA, 44],
            "en_list": ["list1", "list1", "list1", "list1"],
            "en_1": [pd.NA, "node1", "node2", "node3"],
            "en_2": [pd.NA, pd.NA, pd.NA, pd.NA],
        }
    )
    return ExcelSheet(excel_name="file1", sheet_name="sheet1", df=df, col_info=cols_en_1_2)


@pytest.fixture
def f1_s1_no_list_columns(cols_en_list_only: Columns) -> ExcelSheet:
    df = pd.DataFrame({"en_list": ["list1", "list1", "list1", "list1"]})
    return ExcelSheet(excel_name="file1", sheet_name="sheet1", df=df, col_info=cols_en_list_only)


@pytest.fixture
def f1_s1_empty_list_column(cols_en_1_2: Columns) -> ExcelSheet:
    df = pd.DataFrame(
        {
            "id (optional)": [11, 22, pd.NA, 44],
            "en_list": [pd.NA, pd.NA, pd.NA, pd.NA],
            "en_1": [pd.NA, "node1", "node2", "node3"],
        }
    )
    return ExcelSheet(excel_name="file1", sheet_name="sheet1", df=df, col_info=cols_en_1_2)


@pytest.fixture
def f1_s1_missing_translation_id_filled(cols_en_de_1_2_no_comment: Columns) -> ExcelSheet:
    df = pd.DataFrame(
        {
            "id": ["list_id", "1", "1.1", "2", "3", "3.1", "3.2", "3.2.1", "3.2.2"],
            "parent_id": ["list_id", "list_id", "1", "list_id", "list_id", "3", "3", "3.2", "3.2"],
            "en_list": [
                "Listname_en",
                "Listname_en",
                "Listname_en",
                "Listname_en",
                "Listname_en",
                pd.NA,
                "Listname_en",
                "Listname_en",
                "Listname_en",
            ],
            "de_list": [
                "Listname_de",
                "Listname_de",
                "Listname_de",
                "Listname_de",
                "Listname_de",
                "Listname_de",
                "Listname_de",
                "Listname_de",
                "Listname_de",
            ],
            "en_1": [
                pd.NA,
                "Node_en_1",
                "Node_en_1",
                "Node_en_2",
                "Node_en_3",
                "Node_en_3",
                "Node_en_3",
                "Node_en_3",
                pd.NA,
            ],
            "de_1": [
                pd.NA,
                "Node_de_1",
                "Node_de_1",
                "Node_de_2",
                "Node_de_3",
                "Node_de_3",
                "Node_de_3",
                "Node_de_3",
                "Node_de_3",
            ],
            "en_2": [
                pd.NA,
                pd.NA,
                "Node_en_1.1",
                pd.NA,
                pd.NA,
                "Node_en_3.1",
                "Node_en_3.2",
                "Node_en_3.2",
                "Node_en_3.2",
            ],
            "de_2": [
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                "Node_de_3.1",
                "Node_de_3.2",
                "Node_de_3.2",
                "Node_de_3.2",
            ],
        }
    )
    return ExcelSheet(excel_name="excel1", sheet_name="sheet1", df=df, col_info=cols_en_de_1_2_no_comment)


@pytest.fixture
def f1_s2_list_empty() -> ExcelSheet:
    df = pd.DataFrame({"en_list": [pd.NA, pd.NA], "en_1": ["list1", "list1"]})
    return ExcelSheet(
        excel_name="file1",
        sheet_name="sheet2",
        df=df,
        col_info=Columns(
            preferred_lang="en", list_cols=["en_list"], comment_cols=[], node_cols=[ColumnNodes(1, ["en_1"])]
        ),
    )


@pytest.fixture
def f1_s1_no_nodes() -> ExcelSheet:
    df = pd.DataFrame({"en_list": ["list1", "list1"]})
    return ExcelSheet(
        excel_name="file1",
        sheet_name="sheet1",
        df=df,
        col_info=Columns(preferred_lang="en", list_cols=["en_list"], comment_cols=[], node_cols=[]),
    )


@pytest.fixture
def f2_s2_missing_lang_column(cols_en_de_1_3: Columns) -> ExcelSheet:
    df = pd.DataFrame(
        {
            "en_list": ["list1", "list1", "list1", "list1", "list1", "list1", "list1", "list1"],
            "de_list": ["list1", "list1", "list1", "list1", "list1", "list1", "list1", "list1"],
            "de_comments": ["comment de list", pd.NA, pd.NA, "comment de node1.1.1", pd.NA, pd.NA, pd.NA, pd.NA],
            "en_1": [pd.NA, "node1", "node1", "node1", "node1", "node2", "node2", "node3"],
            "de_1": [pd.NA, "node1", "node1", "node1", "node1", "node2", "node2", "node3"],
            "en_2": [pd.NA, pd.NA, "node1.1", "node1.1", "node1.2", pd.NA, "node2.1", pd.NA],
            "de_2": [pd.NA, pd.NA, "node1.1", "node1.1", "node1.2", pd.NA, "node2.1", pd.NA],
            "en_3": [pd.NA, pd.NA, pd.NA, "node1.1.1", pd.NA, pd.NA, pd.NA, pd.NA],
        }
    )
    return ExcelSheet(excel_name="file2", sheet_name="sheet2", df=df, col_info=cols_en_de_1_3)


@pytest.fixture
def f2_s2_missing_translations(cols_en_de_1_3: Columns) -> ExcelSheet:
    df = pd.DataFrame(
        {
            "id (optional)": [1, 2, 3, 4, 5, 6, 7],
            "en_list": ["list3", "list3", "list3", "list3", pd.NA, "list3", "list3"],
            "de_list": [pd.NA, "list3", "list3", "list3", "list3", "list3", "list3"],
            "en_comments": ["comment en list", pd.NA, pd.NA, "comment en node1.1.1", pd.NA, pd.NA, pd.NA],
            "de_comments": ["comment de list", pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA],
            "en_1": [pd.NA, pd.NA, "node1", "node1", pd.NA, "node1", "node3"],
            "de_1": [pd.NA, "node1", "node1", "node1", "node1", "node2", "node3"],
            "en_2": [pd.NA, pd.NA, "node1.1", "node1.1", pd.NA, pd.NA, pd.NA],
            "de_2": [pd.NA, pd.NA, "node1.1", "node1.1", "node1.2", pd.NA, pd.NA],
            "en_3": [pd.NA, pd.NA, pd.NA, "node1.1.1", pd.NA, pd.NA, pd.NA],
            "de_3": [pd.NA, pd.NA, pd.NA, "node1.1.1", pd.NA, pd.NA, pd.NA],
        }
    )
    return ExcelSheet(excel_name="file2", sheet_name="sheet2", df=df, col_info=cols_en_de_1_3)


@pytest.fixture
def f2_s3_one_row(cols_en_1: Columns) -> ExcelSheet:
    df = pd.DataFrame(
        {
            "en_list": ["list1"],
            "en_1": ["node1"],
        }
    )
    return ExcelSheet(excel_name="file2", sheet_name="sheet3", df=df, col_info=cols_en_1)


@pytest.fixture
def f1_s1_identical_row(cols_en_1: Columns) -> ExcelSheet:
    df = pd.DataFrame(
        {
            "en_list": ["list1", "list1", "list1", "list1"],
            "en_1": [pd.NA, "node1", "node1", "node3"],
            "id (optional)": [1, 2, 3, 4],
        }
    )
    return ExcelSheet(excel_name="file1", sheet_name="sheet1", df=df, col_info=cols_en_1)


@pytest.fixture
def sheets_duplicate_id(cols_en_1: Columns) -> list[ExcelSheet]:
    df_1 = pd.DataFrame(
        {
            "en_list": ["list1", "list1", "list1", "list1"],
            "en_1": [pd.NA, "node1", "node2", "node3"],
            "id (optional)": [2, 1, 3, 4],
        }
    )
    df_2 = pd.DataFrame(
        {
            "en_list": ["list2", "list2", "list2"],
            "en_1": [pd.NA, "node1", "node3"],
            "id (optional)": [1, 22, 4],
        }
    )
    return [
        ExcelSheet(excel_name="file1", sheet_name="sheet1", df=df_1, col_info=cols_en_1),
        ExcelSheet(excel_name="file2", sheet_name="sheet2", df=df_2, col_info=cols_en_1),
    ]



from __future__ import annotations

import warnings
from dataclasses import dataclass
from typing import Any
from typing import Protocol

from lxml import etree

from dsp_tools.models.custom_warnings import DspToolsUserWarning
from dsp_tools.utils.uri_util import is_uri
from dsp_tools.xmllib.value_checkers import is_bool_like
from dsp_tools.xmllib.value_checkers import is_color
from dsp_tools.xmllib.value_checkers import is_date
from dsp_tools.xmllib.value_checkers import is_decimal
from dsp_tools.xmllib.value_checkers import is_geoname
from dsp_tools.xmllib.value_checkers import is_integer
from dsp_tools.xmllib.value_checkers import is_string_like
from dsp_tools.xmllib.value_checkers import is_timestamp
from dsp_tools.xmllib.value_converters import convert_to_bool_string

XML_NAMESPACE_MAP = {None: "https://dasch.swiss/schema", "xsi": "http://www.w3.org/2001/XMLSchema-instance"}
DASCH_SCHEMA = "{https://dasch.swiss/schema}"


class Value(Protocol):
    value: Any
    prop_name: str
    permissions: str | None
    comment: str | None
    resource_id: str | None

    def serialise(self) -> etree._Element:
        raise NotImplementedError

    def make_prop(self) -> etree._Element:
        raise NotImplementedError

    def make_element(self) -> etree._Element:
        raise NotImplementedError


@dataclass
class BooleanValue:
    value: Any
    prop_name: str
    permissions: str | None = "prop-default"
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_bool_like(self.value):
            _warn_type_mismatch(
                expected_type="bool", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )
        self.value = convert_to_bool_string(self.value)

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}boolean-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {"permissions": self.permissions} if self.permissions else {}
        if self.comment:
            attribs["comment"] = self.comment
        ele = etree.Element(f"{DASCH_SCHEMA}boolean", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = str(self.value)
        return ele


@dataclass
class ColorValue:
    value: Any
    prop_name: str
    permissions: str | None = "prop-default"
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_color(self.value):
            _warn_type_mismatch(
                expected_type="color", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}color-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {"permissions": self.permissions} if self.permissions else {}
        if self.comment:
            attribs["comment"] = self.comment
        ele = etree.Element(f"{DASCH_SCHEMA}color", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = str(self.value)
        return ele


@dataclass
class DateValue:
    value: str
    prop_name: str
    permissions: str | None = "prop-default"
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_date(self.value):
            _warn_type_mismatch(
                expected_type="date", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}date-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {"permissions": self.permissions} if self.permissions else {}
        if self.comment:
            attribs["comment"] = self.comment
        ele = etree.Element(f"{DASCH_SCHEMA}date", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = str(self.value)
        return ele


@dataclass
class DecimalValue:
    value: Any
    prop_name: str
    permissions: str | None = "prop-default"
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_decimal(self.value):
            _warn_type_mismatch(
                expected_type="decimal", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}decimal-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {"permissions": self.permissions} if self.permissions else {}
        if self.comment:
            attribs["comment"] = self.comment
        ele = etree.Element(f"{DASCH_SCHEMA}decimal", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = str(self.value)
        return ele


@dataclass
class GeonameValue:
    value: Any
    prop_name: str
    permissions: str | None = "prop-default"
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_geoname(self.value):
            _warn_type_mismatch(
                expected_type="geoname", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}geoname-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {"permissions": self.permissions} if self.permissions else {}
        if self.comment:
            attribs["comment"] = self.comment
        ele = etree.Element(f"{DASCH_SCHEMA}geoname", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = str(self.value)
        return ele


@dataclass
class IntValue:
    value: int | str
    prop_name: str
    permissions: str | None = "prop-default"
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_integer(self.value):
            _warn_type_mismatch(
                expected_type="integer", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}integer-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {"permissions": self.permissions} if self.permissions else {}
        if self.comment:
            attribs["comment"] = self.comment
        ele = etree.Element(f"{DASCH_SCHEMA}integer", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = str(self.value)
        return ele


@dataclass
class LinkValue:
    value: str
    prop_name: str
    permissions: str | None = "prop-default"
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_string_like(self.value):
            _warn_type_mismatch(
                expected_type="string", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}resptr-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {"permissions": self.permissions} if self.permissions else {}
        if self.comment:
            attribs["comment"] = self.comment
        ele = etree.Element(f"{DASCH_SCHEMA}resptr", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = self.value
        return ele


@dataclass
class ListValue:
    value: Any
    list_name: Any
    prop_name: str
    permissions: str | None = "prop-default"
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_string_like(self.value) or not is_string_like(self.list_name):
            _warn_type_mismatch(
                expected_type="list", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(
            f"{DASCH_SCHEMA}list-prop", name=self.prop_name, list=self.list_name, nsmap=XML_NAMESPACE_MAP
        )

    def make_element(self) -> etree._Element:
        attribs = {"permissions": self.permissions} if self.permissions else {}
        if self.comment:
            attribs["comment"] = self.comment
        ele = etree.Element(f"{DASCH_SCHEMA}list", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = self.value
        return ele


@dataclass
class SimpleText:
    value: str
    prop_name: str
    permissions: str | None = "prop-default"
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_string_like(self.value):
            _warn_type_mismatch(
                expected_type="string", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}text-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {"encoding": "utf8"}
        if self.permissions:
            attribs["permissions"] = self.permissions
        if self.comment:
            attribs["comment"] = self.comment
        ele = etree.Element(f"{DASCH_SCHEMA}text", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = self.value
        return ele


@dataclass
class Richtext:
    value: str
    prop_name: str
    permissions: str | None = "prop-default"
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_string_like(self.value):
            _warn_type_mismatch(
                expected_type="string", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}text-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {"encoding": "xml"}
        if self.permissions:
            attribs["permissions"] = self.permissions
        if self.comment:
            attribs["comment"] = self.comment
        ele = etree.Element(f"{DASCH_SCHEMA}text", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = self.value
        return ele


@dataclass
class TimeValue:
    value: str
    prop_name: str
    permissions: str | None = "prop-default"
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_timestamp(self.value):
            _warn_type_mismatch(
                expected_type="timestamp", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}time-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {"permissions": self.permissions} if self.permissions else {}
        if self.comment:
            attribs["comment"] = self.comment
        ele = etree.Element(f"{DASCH_SCHEMA}time", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = self.value
        return ele


@dataclass
class UriValue:
    value: str
    prop_name: str
    permissions: str | None = "prop-default"
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_uri(self.value):
            _warn_type_mismatch(
                expected_type="uri", value=self.value, prop_name=self.prop_name, res_id=self.resource_id
            )

    def serialise(self) -> etree._Element:
        ele = self.make_prop()
        ele.append(self.make_element())
        return ele

    def make_prop(self) -> etree._Element:
        return etree.Element(f"{DASCH_SCHEMA}uri-prop", name=self.prop_name, nsmap=XML_NAMESPACE_MAP)

    def make_element(self) -> etree._Element:
        attribs = {"permissions": self.permissions} if self.permissions else {}
        if self.comment:
            attribs["comment"] = self.comment
        ele = etree.Element(f"{DASCH_SCHEMA}uri", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = self.value
        return ele


def _warn_type_mismatch(expected_type: str, value: Any, prop_name: str, res_id: str | None) -> None:
    """Emits a warning if a values is not in the expected format."""
    msg = f"At the following location a '{expected_type}' does not conform to the expected format.\n"
    msg += f"Resource: {res_id} | " if res_id else ""
    msg += f"Value: {value} | Property: {prop_name}"
    warnings.warn(DspToolsUserWarning(msg))



from dsp_tools.commands.excel2xml.excel2xml_lib import _escape_reserved_chars


class TestEscapedChars:
    def test_single_br(self) -> None:
        test_text = "Text <br/> text after"
        res = _escape_reserved_chars(test_text)
        assert res == test_text

    def test_single_br_with_other(self) -> None:
        test_text = "Text <br/>> text after"
        expected = "Text <br/>&gt; text after"
        res = _escape_reserved_chars(test_text)
        assert res == expected

    def test_wrong_single_br(self) -> None:
        test_text = "Text <br//> text after"
        expected = "Text &lt;br//&gt; text after"
        res = _escape_reserved_chars(test_text)
        assert res == expected

    def test_emphasis(self) -> None:
        test_text = "Text before [<em>emphasis</em>] Text after illegal amp: &"
        expected = "Text before [<em>emphasis</em>] Text after illegal amp: &amp;"
        res = _escape_reserved_chars(test_text)
        assert res == expected

    def test_link(self) -> None:
        test_text = 'Before <a class="salsah-link" href="IRI:link:IRI">link</a> after'
        res = _escape_reserved_chars(test_text)
        assert res == test_text

    def test_illegal_angular(self) -> None:
        test_text = "Before <TagNotKnown>in tags</TagNotKnown> After."
        expected = "Before &lt;TagNotKnown&gt;in tags&lt;/TagNotKnown&gt; After."
        res = _escape_reserved_chars(test_text)
        assert res == expected
